import ujson
import os

from edge import eventlog
from edge.service.jnode.types import JsonGpsdServiceArgs, JsonTzServiceArgs, TzServiceArgs, GpsdServiceArgs
from edge.image.jnode.types import JsonImFooterArgs, ImFooterArgs


class StatServiceArgs:
    def __init__(self, stat_filepath, stat_tout_sec):
        """
        :type stat_filepath: str
        :type stat_tout_sec: int
        """
        self.stat_filepath = stat_filepath
        self.stat_tout_sec = stat_tout_sec

    @staticmethod
    def deserialize(monitoring_jnode):
        """
        :type monitoring_jnode: dict
        :rtype: StatServiceArgs
        :raise JnodeError:
        """
        stat_filepath = monitoring_jnode.get('path')
        if not (type(stat_filepath) in [str, unicode] and len(stat_filepath) < 64 and os.path.isabs(stat_filepath)):
            eventlog.log_i('fail: main.conf#/store/path')
            raise ValueError('main.conf#/store/path')

        stat_tout_sec = monitoring_jnode.get('tout')
        if not (int == type(stat_tout_sec) and 0 <= stat_tout_sec <= 255):
            eventlog.log_i('fail: main.conf#/store/tout')
            raise ValueError('main.conf#/store/tout')

        return StatServiceArgs(stat_filepath, stat_tout_sec)


class ImageQueueArgs:
    def __init__(self, endpoint, min_file_count, max_file_count):
        """
        :type endpoint: str
        :type min_file_count: int
        :type max_file_count: int
        """
        self.endpoint = endpoint
        self.min_file_count = min_file_count
        self.max_file_count = max_file_count


class JsonQueueArgs:
    def __init__(self, endpoint, min_file_count, max_file_count):
        """
        :type endpoint: str
        :type min_file_count: int
        :type max_file_count: int
        """
        self.endpoint = endpoint
        self.min_file_count = min_file_count
        self.max_file_count = max_file_count


class SceneReactArgs:
    def __init__(self, log_level, endpoint, dh):
        """
        :type log_level: str
        :type endpoint: str
        """
        self.log_level = log_level
        self.react_dir = endpoint
        self.image_wh = (None, None)
        self.dh = dh

    @staticmethod
    def unpack(log_level, jnode, dh):
        """
        :type log_level: str

        :param jnode: main.conf#/Source
        :type jnode: dict
        :type dh: int
        :rtype: SceneReactArgs
        """
        endpoint = jnode['Url']
        return SceneReactArgs(log_level, endpoint, dh)


class ImageStoreArgs:
    _FORMATS = ['JPEG', 'RGBA']

    def __init__(self, im_format, im_queue_args):
        """
        :type im_format: str
        :type im_queue_args: ImageQueueArgs

        """
        self.codec = im_format
        self.image_queue_args = im_queue_args

    @staticmethod
    def unpack(jnode):
        """
        :param jnode: main.conf#/image-store
        :type jnode: dict
        :rtype: _ImageResizer
        """
        im_max_count = jnode.get('max-file-count', 4)

        im_format = jnode['format']
        if im_format not in ImageStoreArgs._FORMATS:
            eventlog.log_e("Codec value is not valid")

        if not (int == type(im_max_count) and 4 <= im_max_count <= 8):
            raise ValueError('main.conf#/event-store/max-file-count')

        image_store_dir = jnode['endpoint']
        if not (unicode == type(image_store_dir) and len(image_store_dir) < 64 and os.path.isabs(image_store_dir)):
            raise ValueError('main.conf#/store/endpoint')

        image_queue_args = ImageQueueArgs(image_store_dir, im_max_count - 2, im_max_count)

        return ImageStoreArgs(im_format, image_queue_args)


class SceneEventArgs:
    def __init__(self, json_queue_args_list):
        """
        :type json_queue_args_list: list of JsonQueueArgs
        """
        self.json_queue_args_list = json_queue_args_list
        self.log_level = 'info'

    @staticmethod
    def deserialize(scene_event_jnode):
        """
        :type scene_event_jnode: dict
        :rtype: SceneEventArgs
        """
        max_file_count = scene_event_jnode.get('max-file-count', 4)

        if not (int == type(max_file_count) and 4 <= max_file_count <= 8):
            raise ValueError('main.conf#/event-store/max-file-count')

        json_queue_args_list = map(
            lambda scene_endpoint: JsonQueueArgs(scene_endpoint, max_file_count - 5, max_file_count),
            scene_event_jnode['endpoints'])

        scene_event = SceneEventArgs(json_queue_args_list)
        return scene_event


class LrAnnoArgs:
    _MODE_ALWAYS = 'always'
    _MODE_NEVER = 'never'

    def __init__(self, dir_mode, lane_mode, text_mode, trace_mode, margin_mode):
        """
        :type dir_mode: str
        :type lane_mode: str
        :type text_mode: str
        :type trace_mode: str
        :type margin_mode: str
        """
        self.show_dir = dir_mode == 'always'
        self.show_lane = lane_mode == 'always'
        self.show_text = text_mode == 'always'
        self.show_trace = trace_mode == 'always'
        self.show_margin = margin_mode == 'always'


class RaAnnoArgs:
    _MODE_IF_PLATE = 'if-plate'
    _MODE_ALWAYS = 'always'
    _MODE_NEVER = 'never'

    def __init__(self, len_mode, dir_mode, pos_mode, speed_mode, yellow_speed, red_speed):
        """
        :type speed_mode: str
        :type len_mode: str
        :type dir_mode: str
        :type pos_mode: str
        :type yellow_speed: int
        :type red_speed: int
        """
        self._speed_mode = speed_mode
        self._len_mode = len_mode
        self._dir_mode = dir_mode
        self._pos_mode = pos_mode

        self.yellow_speed = yellow_speed
        self.red_speed = red_speed

    @classmethod
    def _show(cls, mode, lr_ra_match):
        """
        :type lr_ra_match: bool
        :rtype: bool
        """
        if mode == cls._MODE_ALWAYS:
            return True

        elif mode == cls._MODE_IF_PLATE:
            return lr_ra_match

        else:
            return False

    def show_len(self, lr_ra_match=False):
        """
        :type lr_ra_match: bool
        :rtype: bool
        """
        return self._show(self._len_mode, lr_ra_match)

    def show_dir(self, lr_ra_match=False):
        """
        :type lr_ra_match: bool
        :rtype: bool
        """
        return self._show(self._dir_mode, lr_ra_match)

    def show_pos(self, lr_ra_match=False):
        """
        :type lr_ra_match: bool
        :rtype: bool
        """
        return self._show(self._pos_mode, lr_ra_match)

    def show_speed(self, lr_ra_match=False):
        """
        :type lr_ra_match: bool
        :rtype: bool
        """
        return self._show(self._speed_mode, lr_ra_match)


class RendererArgs:
    _LR = ["always", "never"]
    _RA = ["always", "never", "if-plate"]

    def __init__(self, lr_anno, ra_anno):
        """
        :type lr_anno: LrAnnoArgs
        :type ra_anno: RaAnnoArgs
        """
        self.lr_anno = lr_anno
        self.ra_anno = ra_anno

    @staticmethod
    def deserialize(renderer_jnode, yellow_speed, red_speed):
        """
        :type renderer_jnode: dict
        :type yellow_speed: int
        :type red_speed: int
        :rtype: RendererArgs
        """

        # lr anno
        lr_jnode = renderer_jnode.get('lr')
        if lr_jnode is not None:
            # lr1
            video_plate = 'never'
            lr_text_mode = lr_jnode.get('plate', video_plate)
            if lr_text_mode not in RendererArgs._LR:
                lr_text_mode = video_plate

            # lr2
            video_plate_trace = 'never'
            lr_trace_mode = lr_jnode.get('trace', video_plate_trace)
            if lr_trace_mode not in RendererArgs._LR:
                lr_trace_mode = video_plate_trace

            # lr3
            video_body_image = 'never'
            lr_margin_mode = lr_jnode.get('body', video_body_image)
            if lr_margin_mode not in RendererArgs._LR:
                lr_margin_mode = video_plate_trace

            # lr4
            video_dir = 'never'
            lr_dir_mode = lr_jnode.get('dir', video_dir)
            if lr_dir_mode not in RendererArgs._LR:
                lr_dir_mode = video_dir

            # lr5
            video_lane = 'never'
            lr_lane_mode = lr_jnode.get('lane', video_lane)
            if lr_lane_mode not in RendererArgs._LR:
                lr_lane_mode = video_lane

        else:
            lr_dir_mode = lr_lane_mode = lr_text_mode = lr_trace_mode = lr_margin_mode = 'never'

        # ra anno
        ra_jnode = renderer_jnode.get('ra')
        if ra_jnode is not None:
            # ra1
            radar_position = 'never'
            ra_pos_mode = ra_jnode.get('pos', radar_position)
            if ra_pos_mode not in RendererArgs._RA:
                ra_pos_mode = radar_position

            # ra2
            radar_speed = 'never'
            ra_speed_mode = ra_jnode.get('speed', radar_speed)
            if ra_speed_mode not in RendererArgs._RA:
                ra_speed_mode = radar_speed

            # ra3
            radar_dir = 'never'
            ra_dir_mode = ra_jnode.get('dir', radar_dir)
            if ra_dir_mode not in RendererArgs._RA:
                ra_dir_mode = radar_dir

            # ra4
            radar_length = 'never'
            ra_len_mode = ra_jnode.get('len', radar_length)
            if ra_len_mode not in RendererArgs._RA:
                ra_len_mode = radar_length

        else:
            ra_len_mode = ra_dir_mode = ra_pos_mode = ra_speed_mode = 'never'

        lr_anno = LrAnnoArgs(lr_dir_mode, lr_lane_mode, lr_text_mode, lr_trace_mode, lr_margin_mode)
        ra_anno = RaAnnoArgs(ra_len_mode, ra_dir_mode, ra_pos_mode, ra_speed_mode, yellow_speed, red_speed)

        return RendererArgs(lr_anno, ra_anno)


class LaneDetectorArgs:
    def __init__(self, lane_to_mask_map, img_width, img_height):
        """
        :type lane_to_mask_map: list[(int, list[(int, int)])] | list[(str, list[(int, int)])]
        :type img_width: int
        :type img_height: int
        """
        self.id_to_lane_list = lane_to_mask_map
        self.img_width = img_width
        self.img_height = img_height

    @staticmethod
    def unpack(jnode):
        """
        :param jnode: main.conf#/road-mask
        :type jnode: dict
        :rtype: LaneDetectorArgs
        """
        lanes_jnode = jnode.get('lanes', [])
        lane_to_mask_list = map(lambda lane_jnode: (lane_jnode['name'], lane_jnode['mask']), lanes_jnode)
        img_w, img_h = jnode.get('scene-wh', [0, 0])

        return LaneDetectorArgs(lane_to_mask_list, img_w, img_h)


class KernelArgs:
    def __init__(self, add_scene_react, scene_event, image_store, renderer, lane_detector_args, lr_store_tout, tz, gpsd,
                 footer, monitoring):
        """
        :type add_scene_react: SceneReactArgs
        :type scene_event: SceneEventArgs
        :type image_store: ImageStoreArgs
        :type renderer: RendererArgs
        :type lane_detector_args: LaneDetectorArgs
        :type lr_store_tout: int
        :type tz: TzServiceArgs
        :type gpsd: GpsdServiceArgs
        :type footer: ImFooterArgs
        :type monitoring: StatServiceArgs
        """
        self.lane_detector_args = lane_detector_args
        self.add_scene_react = add_scene_react
        self.scene_event = scene_event
        self.image_store = image_store
        self.renderer = renderer
        self.lr_store_tout = lr_store_tout
        self.tz = tz
        self.gpsd = gpsd
        self.footer = footer
        self.monitoring = monitoring

    @staticmethod
    def unpack(jnode, log_level):
        """
        :param jnode: main.conf#/
        :type jnode: dict
        :type log_level: str
        :rtype: KernelArgs
        """
        lane_detector = LaneDetectorArgs.unpack(jnode.get('road-mask', {}))

        image_store = ImageStoreArgs.unpack(jnode['image-store'])

        renderer_jnode = jnode['Renderer']
        profile_jnode = renderer_jnode['profile']
        footer = JsonImFooterArgs.deserialize(profile_jnode['footer'])

        react = SceneReactArgs.unpack(log_level, jnode['Source'], footer.height)

        scene_event = SceneEventArgs.deserialize(jnode['scene-event'])
        scene_event.log_level = log_level

        imw = profile_jnode['width']
        react.image_wh = (imw, 65535)

        vehicle_jnode = profile_jnode.get('annotation', {}).get('vehicle', {})

        yellow = renderer_jnode.get('YellowSpeedThreshold', 65535)
        red = renderer_jnode.get('RedSpeedThreshold', 65535)
        renderer = RendererArgs.deserialize(vehicle_jnode, yellow, red)

        # define LrStoreTout
        def_lr_store_tout = 1
        lr_store_tout = renderer_jnode.get('LrStoreTout', def_lr_store_tout)
        if not (1 <= lr_store_tout <= 60):

            eventlog.log_w('The param LrStoreTout has invalid value %d. The default value %d is used.'
                                 % (lr_store_tout, def_lr_store_tout))
            lr_store_tout = def_lr_store_tout

        # deserialize services_jnode
        services_jnode = jnode['services']

        # datetime service
        tz_conf_filepath = services_jnode['datetime']
        with open(tz_conf_filepath) as tz_conf_jnode:
            tz_conf_jnode = ujson.load(tz_conf_jnode)

        tz = JsonTzServiceArgs.deserialzie(tz_conf_jnode)

        # geolocation service
        gps_conf_filepath = services_jnode['geolocation']
        with open(gps_conf_filepath) as geolocation_jnode:
            geolocation_jnode = ujson.load(geolocation_jnode)

        gpsd = JsonGpsdServiceArgs.deserialzie(geolocation_jnode)

        # stat service
        monitoring_jnode = services_jnode['monitoring']
        monitoring = StatServiceArgs.deserialize(monitoring_jnode)

        return KernelArgs(react, scene_event, image_store, renderer, lane_detector, lr_store_tout, tz, gpsd, footer,
                          monitoring)
