
from _event import SceneEventConsumer
from _renderer import Renderer
from _color_store import LrColorToutStore

from twisted.python.runtime import platform
from edge.lane_detector import LaneDetector
from edge import eventlog
from edge.image.image import ImRgba, LiveImEngine, LiveImArgs

from edge.image.types import ImFooterVars

from edge.service import tz
from edge.service import gpsd


from ._args import KernelArgs
from ._stats import init_service as init_monitoring, fini_service as fini_monitoring, IStatService

if platform.isLinux():
    from _react import INotifyReactProducer as ReactProducer
else:
    from _react import StubReactProducer as ReactProducer


class KernelServices:
    def __init__(self, tz_service, gps_service, stat_service):
        """
        :type tz_service: tz.ITzService
        :type gps_service: gpsd.IGpsdService
        :type stat_service: IStatService
        """
        self.tz = tz_service
        self.gps = gps_service
        self.stat = stat_service


class BaseKernel:
    def __init__(self, react_producer, event_consumer, image_renderer, im_engine, services):
        """
        :type react_producer: ReactProducer
        :type event_consumer: SceneEventConsumer
        :type image_renderer: Renderer
        :type services: KernelServices
        """

        self._react_producer = react_producer
        self._react_producer.set_handler(self.handle_react)
        self._event_consumer = event_consumer
        self._image_renderer = image_renderer

        self._im_engine = im_engine
        self._services = services

        #
        self._react_ts = 0

    def handle_react(self, tr_scene, rgba_image):
        """
        :type tr_scene: _models.TrScene
        :type rgba_image: ImRgba
        """
        # noinspection PyBroadException
        try:
            react_ts = tr_scene.im_ref.ts
            if react_ts > self._react_ts:
                self._react_ts = react_ts

                geodata = self._services.gps.get_geodata(rgba_image.ts)
                f_vars = ImFooterVars({}, rgba_image.ts, None, geodata, None, None, self._services.tz)
                self._image_renderer.render_tr_scene(tr_scene, rgba_image)
                self._im_engine.mk_live(rgba_image, f_vars)
                self._event_consumer.consume_scene(react_ts, rgba_image)
                self._services.stat.wh.im_size(rgba_image.w, rgba_image.h, rgba_image.dh)

        except Exception:
            eventlog.log_x()


class Kernel(BaseKernel):
    def __init__(self, twisted_reactor, args):
        """
        :type twisted_reactor: twisted.internet.interfaces.IReactorTCP
        :type args: KernelArgs
        """

        lane_detector = LaneDetector(
            args.lane_detector_args.id_to_lane_list,
            args.lane_detector_args.img_width,
            args.lane_detector_args.img_height)

        react_producer = ReactProducer(
            twisted_reactor,
            args.add_scene_react)

        tz_service = tz.init_service(args.tz)
        gpsd_service = gpsd.init_service(twisted_reactor, args.gpsd)
        stat_service = init_monitoring(twisted_reactor, args.monitoring)

        services = KernelServices(tz_service, gpsd_service, stat_service)

        event_consumer = SceneEventConsumer.create(twisted_reactor, args.scene_event, args.image_store)

        color_tout_store = LrColorToutStore(twisted_reactor, args.lr_store_tout)

        renderer = Renderer(
            args.renderer,
            color_tout_store,
            lane_detector)

        live_args = LiveImArgs(args.footer)
        im_engine = LiveImEngine(live_args)

        BaseKernel.__init__(self, react_producer, event_consumer, renderer, im_engine, services)

        self._twisted_reactor = twisted_reactor

        def shutdown():
            gpsd.fini_service(gpsd_service)
            tz.fini_service(tz_service)
            fini_monitoring(stat_service)

        twisted_reactor.addSystemEventTrigger('before', 'shutdown', shutdown)

    def run(self):
        self._twisted_reactor.run()
