import ujson
from twisted.internet import inotify
from twisted.python.filepath import FilePath

from _models import TrScene
from _args import SceneReactArgs

from edge import eventlog
from edge.com import fs
from edge.image.image import ImLoader, ImRgba


def _listen_react_dir(react_dir, react_handler):
    fs.silent_rm_r(react_dir)
    fs.silent_mkdir_r(react_dir)

    notifier = inotify.INotify()
    notifier.startReading()
    notifier.watch(FilePath(react_dir),
                   mask=inotify.IN_MOVED_TO | inotify.IN_CLOSE_WRITE,
                   callbacks=[react_handler])


def _load_json_file(react_abspath):
    """
    :type react_abspath: str | None
    :rtype: dict
    """
    if not react_abspath:
        return None

    react_jnode = None
    try:
        with open(react_abspath) as json_file:
            react_jnode = ujson.load(json_file)

    except ValueError:
        eventlog.log_x()

    except IOError:
        eventlog.log_x()

    return react_jnode


class IReactProducer:
    def __init__(self):
        pass

    def set_handler(self, handler):
        raise NotImplementedError


# noinspection PyUnusedLocal
def stub_addscene_react_handler(tr_scene, mono8t_image):
    """
    :type tr_scene: _models.TrScene
    :type mono8t_image: PIL.Image.Image
    """
    print 'stub_addscene_react_handler(tr_track, ra_scene)'


class StubReactProducer(IReactProducer):
    # noinspection PyUnusedLocal
    def __init__(self, *args, **kwargs):
        IReactProducer.__init__(self)

    def set_handler(self, handler):
        pass


class INotifyReactProducer(IReactProducer):
    _IM_FORMATS = [
        'MONO8T',
        'I420PLN',
        'NV12PLN',
        'YUV422T',
        'YUV411T',
        'BayerRG8T',
        'BayerGR8T',
        'BayerBG8T',
        'BayerGB8T'
    ]

    def __init__(self, twisted_reactor, args):
        """
        :type args: SceneReactArgs
        """
        IReactProducer.__init__(self)

        self._react_dir = args.react_dir
        self._react_handler = None

        self._twisted_reactor = twisted_reactor
        self._inotify = inotify.INotify()
        self._inotify.startReading()

        self._im_loader = ImLoader(args.dh, args.image_wh[0], args.image_wh[1])
        self._twisted_reactor.addSystemEventTrigger('before', 'shutdown', self._im_loader.destroy)
        self._rgba_image = ImRgba()

    def _reset_inotify_watch(self):
        try:
            self._inotify.ignore(FilePath(self._react_dir))
        except KeyError:
            pass

        fs.silent_mkdir_r(self._react_dir)

        success = False
        try:
            self._inotify.watch(
                FilePath(self._react_dir),
                mask=inotify.IN_MOVED_TO | inotify.IN_CLOSE_WRITE | inotify.IN_DELETE_SELF | inotify.IN_MOVE_SELF,
                callbacks=[self._on_notify])
            success = True
        except inotify.INotifyError:
            pass

        if not success:
            eventlog.log_e('Failed to reset inotify watch')

    def set_handler(self, handler):
        self._react_handler = handler
        self._reset_inotify_watch()

    # noinspection PyUnusedLocal
    def _on_notify(self, ignored, path, mask):
        if 0 != mask & (inotify.IN_DELETE_SELF | inotify.IN_MOVE_SELF):
            self._twisted_reactor.callLater(0, self._reset_inotify_watch)
            eventlog.log_w('React dir was changed to %s' % self._react_dir)
            return

        # noinspection PyBroadException
        try:
            tr_scene_jnode = _load_json_file(path.path)
            tr_scene = TrScene.unpack(tr_scene_jnode)

            im_ref = tr_scene.im_ref
            if im_ref.codec in self._IM_FORMATS:
                try:
                    self._im_loader.load_ref(im_ref, self._rgba_image)
                except ValueError as x:
                    eventlog.log_d('fail: load image %s ' % x.message)
                    raise

            else:
                raise NotImplementedError

            self._react_handler(tr_scene, self._rgba_image)
        except Exception as x:
            eventlog.log_d('fail: handle tr_scene %s ' % x.message)

        finally:
            fs.silent_rm_r(path.path)
