import ujson
from edge import eventlog
from edge.admin.ui_types import *
from edge.path_helpers import *
import edge.pipeline_helpers as helpers
from os.path import join as join_path


def im_try_get_creator_ctor(ui_entry_point, used_ui_modules, used_edge_modules, avaliable_ui_modules):
    """
    :type ui_entry_point: UiIdentity
    :type used_ui_modules: set of UiIdentity
    :type used_edge_modules: set of EdgeIdentity
    :type avaliable_ui_modules: set of UiIdentity
    :rtype: None | (str, UiIdentity, dict of UiIdentity: dict) -> BaseCreator
    """
    creator_ctor = None
    # speed
    if ui_entry_point in [Pipeline.Speed.Ui.LOCATION_0, Pipeline.Speed.Ui.VIDEOSOURCE_0]:

        used_ui_modules.update([Pipeline.Speed.Ui.LOCATION_0, Pipeline.Speed.Ui.VIDEOSOURCE_0])
        used_edge_modules.add(Pipeline.Speed.Edge.IMAGEMAPPER_0)
        creator_ctor = ImCreator.create

    elif ui_entry_point in [Pipeline.Speed.Ui.PIPELINE]:
        used_ui_modules.update([Pipeline.Speed.Ui.LOCATION_0, Pipeline.Speed.Ui.VIDEOSOURCE_0])
        creator_ctor = ImCreator.create

    # mv
    elif ui_entry_point in [Pipeline.Mv.Ui.LOCATION_0, Pipeline.Mv.Ui.VIDEOSOURCE_0]:

        used_ui_modules.update([Pipeline.Mv.Ui.LOCATION_0, Pipeline.Mv.Ui.VIDEOSOURCE_0])
        used_edge_modules.add(Pipeline.Mv.Edge.IMAGEMAPPER_0)
        creator_ctor = ImCreator.create

    elif ui_entry_point in [Pipeline.Mv.Ui.PIPELINE]:
        used_ui_modules.update([Pipeline.Mv.Ui.LOCATION_0, Pipeline.Mv.Ui.VIDEOSOURCE_0])
        creator_ctor = ImCreator.create

    return creator_ctor


class ImCreator(BaseCreator):
    def __init__(self, id_main):
        """
        :type id_main: str, dict
        """
        self._id_main = id_main
        BaseCreator.__init__(self)

    @staticmethod
    def create(ui_pid, ui_entry_point, ui_identity_to_jnode_map):
        """
        :type ui_pid: str
        :type ui_entry_point: UiIdentity
        :type ui_identity_to_jnode_map: dict of UiIdentity: dict
        :rtype: RaCreator
        """
        if ui_entry_point in [Pipeline.Speed.Ui.LOCATION_0, Pipeline.Speed.Ui.PIPELINE, Pipeline.Speed.Ui.VIDEOSOURCE_0,
                              # Pipeline.Rtsp.Ui.LOCATION_0, Pipeline.Rtsp.Ui.PIPELINE, Pipeline.Rtsp.Ui.VIDEOSOURCE_0,
                              Pipeline.Mv.Ui.LOCATION_0, Pipeline.Mv.Ui.PIPELINE, Pipeline.Mv.Ui.VIDEOSOURCE_0]:

            id_main = ImCreator._serialize(ui_pid, ui_entry_point, ui_identity_to_jnode_map)
            creator = ImCreator(id_main)

        else:
            raise NotImplemented('pipeline not supported')

        return creator

    @staticmethod
    def _serialize(ui_pid, ui_entry_point, ui_identity_to_jnode_map):
        """
        :type ui_pid: str
        :type ui_entry_point: UiIdentity
        :type ui_identity_to_jnode_map: dict of UiIdentity: dict | None
        :rtype: str, dict, str
        """
        ui_location_identity = UiIdentity(ui_ptype=ui_entry_point.ptype, ui_type=UI_TYPE_LOCATION, ui_id='0')
        ui_location_jnode = ui_identity_to_jnode_map[ui_location_identity]

        ui_videosource_identity = UiIdentity(ui_ptype=ui_entry_point.ptype, ui_type=UI_TYPE_VIDEOSOURCE, ui_id='0')
        ui_videosource_jnode = ui_identity_to_jnode_map[ui_videosource_identity]

        edge_identity = EdgeIdentity(edge_type=EDGE_TYPE_IMAGE_MAPPER, edge_gid='0')

        edge_group = helpers.get_edge_group(ui_pid)
        edge_id = helpers.get_edge_id(edge_group, edge_identity)

        unit_position_jnode = ui_location_jnode['rw']['unit-position']
        forward = unit_position_jnode['forward']
        left = unit_position_jnode['left']
        up = unit_position_jnode['up']

        # deserialize sensor-rotation
        unit_rotation_manual_jnode = ui_location_jnode['rw']['unit-rotation-manual']
        azimuth = unit_rotation_manual_jnode['azimuth']
        elevation = unit_rotation_manual_jnode['elevation']
        roll = unit_rotation_manual_jnode['roll']

        fov_h, fov_v = ui_videosource_jnode['ro']['fov-hv']

        ui_acquisition_jnode = ui_videosource_jnode['rw']['videostream']
        ui_sensor_jnode = ui_acquisition_jnode['resolution']

        # create main.conf
        main_conf_jnode = {
            "sensor": {
                "dimension": {
                    "fov_h": fov_h,
                    "fov_v": fov_v,
                    "width": ui_sensor_jnode['w'],
                    "height": ui_sensor_jnode['h']
                },
                "position": {
                    "forward": forward,
                    "left": left,
                    "up": up
                },
                "rotation": {
                    "elevation": elevation,
                    "azimuth": azimuth,
                    "pitch": roll
                }
            }
        }

        return edge_id, main_conf_jnode

    def serialize(self):
        edge_id, main_conf_jnode = self._id_main

        edge_jnode_dir = get_edge_jnode_dirpath(EDGE_TYPE_IMAGE_MAPPER, edge_id)
        if not os.path.exists(edge_jnode_dir):
            helpers.silent_mkdir_r(edge_jnode_dir)

        main_conf_filepath = join_path(edge_jnode_dir, 'main.conf')
        with open(main_conf_filepath, 'w') as json_file:
            ujson.dump(main_conf_jnode, json_file, indent=2)


class ImDeleter:
    def __init__(self, edge_im_dirpath):
        """
        :type edge_im_dirpath: str
        """
        self._edge_im_dirpath = edge_im_dirpath

    @staticmethod
    def create(ui_pid, ui_entry_point):
        """
        :type ui_pid: str
        :type ui_entry_point: UiIdentity
        :rtype: RaDeleter
        """
        edge_group = helpers.get_edge_group(ui_pid)

        if ui_entry_point == Pipeline.Mv.Ui.PIPELINE:
            edge_id = helpers.get_edge_id(edge_group, Pipeline.Mv.Edge.IMAGEMAPPER_0)
            edge_type = Pipeline.Mv.Edge.IMAGEMAPPER_0.type
            deleter = ImDeleter(get_edge_jnode_dirpath(edge_type, edge_id))
            return deleter

        elif ui_entry_point == Pipeline.Speed.Ui.PIPELINE:
            edge_id = helpers.get_edge_id(edge_group, Pipeline.Speed.Edge.IMAGEMAPPER_0)
            edge_type = Pipeline.Speed.Edge.IMAGEMAPPER_0.type
            deleter = ImDeleter(get_edge_jnode_dirpath(edge_type, edge_id))
            return deleter

        elif ui_entry_point == Pipeline.Rtsp.Ui.PIPELINE:
            return None

        else:
            raise NotImplementedError

    def run(self):
        helpers.silent_rm_r(self._edge_im_dirpath)