import jsonschema
import ujson

from edge import eventlog
from edge import path_helpers as ph
from edge import pipeline_helpers as oh
from edge.admin import reply_helpers as rh
from edge.admin import ui_types as ui
from edge.com import db, fs

import edge_creator as creator
from ui_pipeline_set import get_pipeline_type


class UiEventStore:
    def __init__(self, handler):
        """
        :type handler: object -> dict
        """
        self._handler = handler

    @staticmethod
    def _check_name_changed(old_ui_jnode, new_ui_jnode):
        return old_ui_jnode['origin-name'] != new_ui_jnode['origin-name']

    @staticmethod
    def _handle_get(ui_jnode_path):
        """
        :type ui_jnode_path: str
        :rtype: dict
        """
        ui_jnode = fs.load_json2(ui_jnode_path)

        ro = ui_jnode.get('ro', dict())
        rw = ui_jnode.get('rw', dict())
        rw['origin-guid'] = ro['uuid']

        return rh.reply_done(rw)

    @staticmethod
    def _handle_put(ui_pid, entry_point, new_ui_jnode):
        """
        :type ui_pid: str
        :type entry_point: ph.UiIdentity
        :type new_ui_jnode: dict

        :rtype: dict
        """
        # load old-ui-jnode
        ui_jnode_filepath = ph.get_ui_pipeline_jnode_filepath(ui_pid, entry_point.type, entry_point.id)
        raw_ui_jnode = fs.load_json2(ui_jnode_filepath)

        if raw_ui_jnode['ro']['uuid'] != new_ui_jnode['origin-guid']:
            eventlog.log_d('fail: bad request (origin-guid is read-only)')
            return rh.reply_fail(rh.REPLY_BAD_REQUEST_400)

        old_ui_jnode = raw_ui_jnode['rw']
        if not UiEventStore._check_name_changed(old_ui_jnode, new_ui_jnode):
            return rh.reply_done(new_ui_jnode)

        def _edit_ui_jnode_handler(ui_identity_to_jnode_map):
            """
            :type ui_identity_to_jnode_map: dict of UiIdentity: dict
            :rtype: bool
            """
            ui_identity_to_jnode_map[entry_point]['rw']['origin-name'] = new_ui_jnode['origin-name']

            origin_uuid = oh.get_origin_uuid(ui_pid)
            oh.edit_db_origin(origin_uuid, name=new_ui_jnode['origin-name'], media=True, manual=True)

            return True

        try:
            creator.generate(ui_pid, entry_point, _edit_ui_jnode_handler)
            return rh.reply_done(new_ui_jnode)

        except db.DbError:
            eventlog.log_d('fail: internal-server-error')
            return rh.reply_fail(rh.REPLY_INTERNAL_SERVER_ERROR_500)

        except ui.LockError:
            return rh.reply_fail(rh.REPLY_CONFLICT_409)

        except Exception as x:
            eventlog.log_d("fail: %s" % x.message)
            return rh.reply_fail(rh.REPLY_INTERNAL_SERVER_ERROR_500)

    @staticmethod
    def _deserialize_get(query_args):
        """
        :type query_args: edge.admin.ui_types.UiQueryArgs
        :rtype: UiRtspStreaming
        """
        if query_args.query_string or query_args.raw_body:
            eventlog.log_d('fail: bad-request')
            return UiEventStore(lambda: rh.reply_fail(rh.REPLY_BAD_REQUEST_400))

        try:
            ui_pid = query_args.get_path_info_item(ui.UI_PIPELINE_PID_INDEX)
            ui_type = query_args.get_path_info_item(ui.UI_PIPELINE_TYPE_INDEX)
            ui_id = query_args.get_path_info_item(ui.UI_PIPELINE_ID_INDEX)
        except IndexError:
            return UiEventStore(lambda: rh.reply_fail(rh.REPLY_NOT_FOUND_404))

        ui_jnode_filepath = ph.get_ui_pipeline_jnode_filepath(ui_pid, ui_type, ui_id)
        return UiEventStore(lambda: UiEventStore._handle_get(ui_jnode_filepath))

    @staticmethod
    def _deserialize_put(query_args):
        """
        :type query_args: edge.admin.ui_types.UiQueryArgs
        :rtype: UiRtspStreaming
        """
        try:
            ui_pid = query_args.get_path_info_item(ui.UI_PIPELINE_PID_INDEX)
            ui_type = query_args.get_path_info_item(ui.UI_PIPELINE_TYPE_INDEX)
            ui_id = query_args.get_path_info_item(ui.UI_PIPELINE_ID_INDEX)
        except IndexError:
            return UiEventStore(lambda: rh.reply_fail(rh.REPLY_NOT_FOUND_404))

        ui_ptype = get_pipeline_type(ui_pid)

        if ui_ptype is None:
            return UiEventStore(lambda: rh.reply_fail(rh.REPLY_NOT_FOUND_404))

        entry_point = ph.UiIdentity(ui_ptype, ui_type, ui_id)

        if query_args.query_string:
            eventlog.log_d('fail: bad-request')
            return UiEventStore(lambda: rh.reply_fail(rh.REPLY_BAD_REQUEST_400))

        # load new-ui-jnode
        new_ui_jnode = ujson.loads(query_args.raw_body) if query_args.raw_body else None
        if not new_ui_jnode:
            eventlog.log_d('fail: bad-request')
            return UiEventStore(lambda: rh.reply_fail(rh.REPLY_BAD_REQUEST_400))

        # load ui-schena-jnode
        try:
            ui_schema_filepath = ph.get_ui_pipeline_schema_filepath(ui_pid, ui_type, ui_id)
            with open(ui_schema_filepath) as json_file:
                ui_schema_jnode = ujson.load(json_file)

        except (IOError, ValueError) as x:
            eventlog.log_d("fail: load ui-schema-jnode (%s)" % x.message)
            return UiEventStore(lambda: rh.reply_fail(rh.REPLY_INTERNAL_SERVER_ERROR_500))

        # validate new-ui-jnode
        try:
            jsonschema.validate(new_ui_jnode, ui_schema_jnode)

        except (ValueError, jsonschema.ValidationError) as x:
            eventlog.log_d("fail: validate new-ui-jnode (%s)" % x.message)
            return UiEventStore(lambda: rh.reply_fail(rh.REPLY_BAD_REQUEST_400))

        # all arguments are valid
        return UiEventStore(lambda: UiEventStore._handle_put(ui_pid, entry_point, new_ui_jnode))

    @staticmethod
    def create(query_args):
        """
        :type query_args: edge.admin.ui_types.UiQueryArgs
        :rtype: UiEventStore
        """
        if query_args.http_method == 'GET':
            return UiEventStore._deserialize_get(query_args)

        elif query_args.http_method == 'PUT':
            return UiEventStore._deserialize_put(query_args)

        else:
            eventlog.log_d("fail: method-not-allowed")
            return UiEventStore(lambda: rh.reply_fail(rh.REPLY_METHOD_NOT_ALLOWED_405))

    def handle(self):
        """
        :rtype: dict
        """
        return self._handler()
