import os
import commands
import jsonschema
import ujson


from edge import eventlog
from edge import path_helpers as ph
from edge import pipeline_helpers as oh
from edge.admin import reply_helpers as rh
from edge.admin import ui_types as ui
from edge.com import fs, net

import edge_creator as creator

from ui_pipeline_set import get_pipeline_type

UI_NETWORK_JNODE_FILEPATH = '/etc/edge/setupstore/admin/actions/network.json'
MANAGEMENT_MAC_FILEPATH = '/etc/network/management-mac'


class UiRtspStreaming:
    TYPE = 'rtsp-streaming'

    def __init__(self, handler):
        """
        :type handler: () -> dict
        """
        self._handler = handler

    @staticmethod
    def _make_rtsp_url(ui_rw_jnode):
        """
        :type ui_rw_jnode: dict
        :rtype: str
        """

        try:
            with open(UI_NETWORK_JNODE_FILEPATH) as json_file:
                ui_network_jnode = ujson.load(json_file)
            mac_address = ui_network_jnode['ro']['mac']

        except IOError as x:
            if x.errno == 2:
                if os.path.exists(MANAGEMENT_MAC_FILEPATH):
                    with open(MANAGEMENT_MAC_FILEPATH, 'r') as mgmt_mac_file:
                        mac_address = mgmt_mac_file.readlines()[0].strip()
                else:
                    raise

            else:
                eventlog.log_d('fail: read network settings')
                raise IOError

        ifname = net.get_ifname(mac_address)
        try:
            ip_addr = commands.getoutput("ip addr show %s | grep -v '127.0.0.1' | grep -w inet | awk '{print $2}'"
                                         % ifname).split('/')[0]
        except Exception:
            eventlog.log_d('fail: discover ip-address')
            raise RuntimeError

        tcp_port = str(ui_rw_jnode['streaming']['tcp-port'])
        ui_rtsp_codec = ui_rw_jnode['live']['videostream']['format']
        if ui_rtsp_codec == 'mjpeg':
            rtsp_codec = 'mjpeg'
        elif ui_rtsp_codec == 'h264-cqp' or 'h264-cbr':
            rtsp_codec = 'h264'
        else:
            eventlog.log_d('unknown rtsp-codec %s' % ui_rtsp_codec)
            raise ValueError

        rtsp_url = 'rtsp://' + ip_addr + ':' + tcp_port + '/live-' + rtsp_codec

        return rtsp_url

    @staticmethod
    def _handle_get(ui_jnode_path, ui_pid, ui_id):
        """
        :type ui_jnode_path: str
        :type ui_pid: str
        :type ui_id: str
        :rtype: dict
        """
        videsource_status = oh.get_videsource_status(ui_pid, ui_id)
        if videsource_status is None:
            eventlog.log_d('fail: get videosource status')
            return rh.reply_fail2(
                rh.REPLY_SERVICE_UNAVAILABLE_503,
                error_code=rh.FAIL_CODE_VIDEOSOURCE_ERROR,
                error_text=rh.FAIL_TEXT_VIDEOSOURCE_ERROR)

        elif videsource_status != 0:
            eventlog.log_d('fail: videosource status error')
            return rh.reply_fail2(
                rh.REPLY_SERVICE_UNAVAILABLE_503,
                error_code=rh.FAIL_CODE_VIDEOSOURCE_ERROR,
                error_text=rh.FAIL_TEXT_VIDEOSOURCE_ERROR)

        ui_jnode = fs.load_json(ui_jnode_path)
        if not ui_jnode:
            eventlog.log_d('fail: load ui-jnode')
            return rh.reply_fail(rh.REPLY_NOT_FOUND_404)

        try:
            rtsp_url = UiRtspStreaming._make_rtsp_url(ui_jnode['rw'])
        except (ValueError, RuntimeError, IOError):
            return rh.reply_fail(rh.REPLY_INTERNAL_SERVER_ERROR_500)

        ui_jnode['rw']['streaming']['url'] = rtsp_url

        return rh.reply_done(ui_jnode['rw'])

    @staticmethod
    def _handle_put(ui_pid, entry_point, new_ui_jnode):
        """
        :type ui_pid: str
        :type entry_point: ph.UiIdentity
        :type new_ui_jnode: dict
        :rtype: dict
        """
        videsource_status = oh.get_videsource_status(ui_pid, entry_point.id)
        if videsource_status is None:
            eventlog.log_d('fail: get videosource status')
            return rh.reply_fail2(
                rh.REPLY_SERVICE_UNAVAILABLE_503,
                error_code=rh.FAIL_CODE_VIDEOSOURCE_ERROR,
                error_text=rh.FAIL_TEXT_VIDEOSOURCE_ERROR)

        elif videsource_status != 0:
            eventlog.log_d('fail: videosource status error')
            return rh.reply_fail2(
                rh.REPLY_SERVICE_UNAVAILABLE_503,
                error_code=rh.FAIL_CODE_VIDEOSOURCE_ERROR,
                error_text=rh.FAIL_TEXT_VIDEOSOURCE_ERROR)

        # load old-ui-jnode
        ui_jnode_filepath = ph.get_ui_pipeline_jnode_filepath(ui_pid, entry_point.type, entry_point.id)
        try:
            with open(ui_jnode_filepath) as json_file:
                raw_ui_jnode = ujson.load(json_file)
                old_ui_jnode = raw_ui_jnode['rw']

        except (IOError, ValueError) as x:
            eventlog.log_d("fail: load old-ui-jnode (%s)" % x.message)
            return rh.reply_fail(rh.REPLY_INTERNAL_SERVER_ERROR_500)

        old_rtsp_url = UiRtspStreaming._make_rtsp_url(old_ui_jnode)
        new_rtsp_url = new_ui_jnode['streaming']['url']

        if old_rtsp_url != new_rtsp_url:
            eventlog.log_d('fail: not allowed to edit rtsp-url')
            return rh.reply_fail(rh.REPLY_BAD_REQUEST_400)

        if old_ui_jnode == new_ui_jnode:
            return rh.reply_done(new_ui_jnode)

        restart_setupstore = []

        def _edit_ui_jnode_handler(ui_identity_to_jnode_map):
            """
            :type ui_identity_to_jnode_map: dict of UiIdentity: dict
            :rtype: bool
            """
            if ui_identity_to_jnode_map[entry_point]['rw']['enabled'] != new_ui_jnode['enabled']:
                restart_setupstore.append(True)

            ui_identity_to_jnode_map[entry_point]['rw'] = new_ui_jnode
            return True

        try:
            creator.generate(ui_pid, entry_point, _edit_ui_jnode_handler)
            if restart_setupstore:
                oh.sighup_setupstore()

            return rh.reply_done(new_ui_jnode)

        except ui.LockError:
            return rh.reply_fail(rh.REPLY_CONFLICT_409)

    @staticmethod
    def _deserialize_get(query_args):
        """
        :type query_args: edge.admin.ui_types.UiQueryArgs
        :rtype: UiRtspStreaming
        """
        try:
            ui_pid = query_args.get_path_info_item(ui.UI_PIPELINE_PID_INDEX)
            ui_type = query_args.get_path_info_item(ui.UI_PIPELINE_TYPE_INDEX)
            ui_id = query_args.get_path_info_item(ui.UI_PIPELINE_ID_INDEX)
        except IndexError:
            return UiRtspStreaming(lambda: rh.reply_fail(rh.REPLY_NOT_FOUND_404))

        ui_ptype = get_pipeline_type(ui_pid)
        if ui_ptype is None:
            return UiRtspStreaming(lambda: rh.reply_fail(rh.REPLY_NOT_FOUND_404))

        if query_args.query_string or query_args.raw_body:
            eventlog.log_d('fail: bad-request')
            return UiRtspStreaming(lambda: rh.reply_fail(rh.REPLY_BAD_REQUEST_400))

        ui_jnode_filepath = ph.get_ui_pipeline_jnode_filepath(ui_pid, ui_type, ui_id)
        return UiRtspStreaming(lambda: UiRtspStreaming._handle_get(ui_jnode_filepath, ui_pid, ui_id))

    @staticmethod
    def _deserialize_put(query_args):
        """
        :type query_args: edge.admin.ui_types.UiQueryArgs
        :rtype: UiRtspStreaming
        """
        try:
            ui_pid = query_args.get_path_info_item(ui.UI_PIPELINE_PID_INDEX)
            ui_type = query_args.get_path_info_item(ui.UI_PIPELINE_TYPE_INDEX)
            ui_id = query_args.get_path_info_item(ui.UI_PIPELINE_ID_INDEX)
        except IndexError:
            return UiRtspStreaming(lambda: rh.reply_fail(rh.REPLY_NOT_FOUND_404))

        ui_ptype = get_pipeline_type(ui_pid)
        if ui_ptype is None:
            return UiRtspStreaming(lambda: rh.reply_fail(rh.REPLY_NOT_FOUND_404))

        entry_point = ph.UiIdentity(ui_ptype, ui_type, ui_id)

        if query_args.query_string:
            eventlog.log_d('fail: unsupported query-string')
            return UiRtspStreaming(lambda: rh.reply_fail(rh.REPLY_BAD_REQUEST_400))

        # load new-ui-jnode
        new_ui_jnode = ujson.loads(query_args.raw_body) if query_args.raw_body else None
        if not new_ui_jnode:
            eventlog.log_d('fail: no new-ui-jnode')
            return UiRtspStreaming(lambda: rh.reply_fail(rh.REPLY_BAD_REQUEST_400))

        # load ui-schema-jnode
        try:
            ui_schema_filepath = ph.get_ui_pipeline_schema_filepath(ui_pid, ui_type, ui_id)
            with open(ui_schema_filepath) as json_file:
                ui_schema_jnode = ujson.load(json_file)

        except (IOError, ValueError) as x:
            eventlog.log_d("fail: load ui-schema-jnode (%s)" % x.message)
            return UiRtspStreaming(lambda: rh.reply_fail(rh.REPLY_INTERNAL_SERVER_ERROR_500))

        # validate new-ui-jnode
        try:
            jsonschema.validate(new_ui_jnode, ui_schema_jnode)

        except (ValueError, jsonschema.ValidationError) as x:
            eventlog.log_d("fail: validate new-ui-jnode (%s)" % x.message)
            return UiRtspStreaming(lambda: rh.reply_fail(rh.REPLY_BAD_REQUEST_400))

        ui_streaming_jnode = new_ui_jnode['streaming']
        if ui_streaming_jnode['transport'] == 'rtp-over-udp-or-tcp':
            if ui_streaming_jnode['min-udp-port'] == 0 and ui_streaming_jnode['max-udp-port'] != 0:
                return UiRtspStreaming(lambda: rh.reply_fail(rh.REPLY_BAD_REQUEST_400))

            if ui_streaming_jnode['max-udp-port'] == 0 and ui_streaming_jnode['min-udp-port'] != 0:
                return UiRtspStreaming(lambda: rh.reply_fail(rh.REPLY_BAD_REQUEST_400))

        # all arguments are valid
        return UiRtspStreaming(lambda: UiRtspStreaming._handle_put(ui_pid, entry_point, new_ui_jnode))

    @staticmethod
    def create(query_args):
        """
        :type query_args: edge.admin.ui_types.UiQueryArgs
        :rtype: UiRtspStreaming
        """
        if query_args.http_method == 'GET':
            return UiRtspStreaming._deserialize_get(query_args)

        elif query_args.http_method == 'PUT':
            return UiRtspStreaming._deserialize_put(query_args)

        else:
            eventlog.log_d("fail: method-not-allowed")
            return UiRtspStreaming(lambda: rh.reply_fail(rh.REPLY_METHOD_NOT_ALLOWED_405))

    def handle(self):
        """
        :rtype: dict
        """
        return self._handler()
