import os
import datetime
import ujson
import uuid

from edge import eventlog
from edge import path_helpers as ph
from edge import pipeline_helpers as helpers
from edge.admin import reply_helpers as rh
from edge.admin import ui_types as ui
from edge.com import fs


import ui_pipeline_set
import edge_creator as creator
from _edge_setupstore import append_ss_group
from ui_pipeline import delete_pipeline, get_pipeline_ext_jnode

_TIME_DEF_JNODE_FILEPATH = '/opt/edge/share/setupstore/ui/time-jnode.json'

_UI_IDENTITY_TO_DATASOURCE_MAP = [
    (ph.Pipeline.Rtsp.Ui.EVENTSTORE_0, False),
    (ph.Pipeline.Rtsp.Ui.GEOLOCATION_0, False),
    (ph.Pipeline.Rtsp.Ui.VIDEOSOURCE_0, True),
    (ph.Pipeline.Rtsp.Ui.LPR_0, False),
    (ph.Pipeline.Rtsp.Ui.LANE_ENFORCEMENT_0, False),
    (ph.Pipeline.Rtsp.Ui.HTTP_STREAMING_0, False)
]

_UI_DETECTORS_TO_INTERNAL_MAP = [
    (ph.Pipeline.Rtsp.Ui.TRANSIT_0, True)
]


class CreatorError(RuntimeError):
    pass


def _create_ui_jnode_set(ui_pid, origin_uuid, detectors_list):
    """
    :type ui_pid: str
    :type origin_uuid: str
    :type detectors_list: list
    :rtype set of UiIdentity
    """

    # read modules def_jnode
    ui_identity_to_def_jnode_map = dict()
    for (ui_identity, _) in _UI_IDENTITY_TO_DATASOURCE_MAP:
        def_ui_jnode_path = ph.get_ui_def_jnode_filepath(ui_identity, None)
        def_ui_jnode = fs.load_json2(def_ui_jnode_path)
        ui_identity_to_def_jnode_map[ui_identity] = def_ui_jnode

    # read detectors def jnode
    for (ui_identity, internal) in _UI_DETECTORS_TO_INTERNAL_MAP:
        def_ui_jnode_path = ph.get_ui_def_jnode_filepath(ui_identity, None)
        def_ui_jnode = fs.load_json2(def_ui_jnode_path)
        ui_identity_to_def_jnode_map[ui_identity] = def_ui_jnode
        detector_def_name = def_ui_jnode['rw'].get('args', {}).get('name', '')
        detectors_list.append((ui_identity, internal, detector_def_name))

    # edit
    ui_identity_to_def_jnode_map[ph.Pipeline.Rtsp.Ui.EVENTSTORE_0]['rw']['channel'] = int(ui_pid)
    ui_identity_to_def_jnode_map[ph.Pipeline.Rtsp.Ui.EVENTSTORE_0]['ro']['uuid'] = origin_uuid
    ui_identity_to_def_jnode_map[ph.Pipeline.Rtsp.Ui.EVENTSTORE_0]['ro']['ctime'] = datetime.datetime.now().isoformat()

    # write
    ui_identity_set = set()
    for ui_identity, ui_jnode in ui_identity_to_def_jnode_map.iteritems():
        ui_identity_set.add(ui_identity)
        ui_jnode_dirpath = ph.get_ui_pipeline_jnode_dirpath(ui_pid, ui_identity)
        if not os.path.exists(ui_jnode_dirpath):
            helpers.silent_mkdir_r(ui_jnode_dirpath)

        ui_jnode_filepath = ph.get_ui_pipeline_jnode_filepath(ui_pid, ui_identity.type, ui_identity.id)
        with open(ui_jnode_filepath, 'w') as json_file:
            ujson.dump(ui_jnode, json_file, indent=2)

        ui_schema_filepath = ph.get_ui_pipeline_schema_filepath(ui_pid, ui_identity.type, ui_identity.id)
        fs.ln_schema(ui_schema_filepath, ui_jnode_filepath)

    return ui_identity_set


def _generator(ui_pid, ui_identity_set):
    """
    :type ui_pid: str
    :type ui_identity_set: set of UiIdentity
    """
    entry_point = ph.Pipeline.Rtsp.Ui.PIPELINE

    try:
        creator.generate(ui_pid, entry_point, None, ui_identity_set)
        return rh.REPLY_SUCCESS_200

    except ui.LockError:
        return rh.REPLY_CONFLICT_409


def create_rtsp_pipeline(ui_pid):
    """
    :type ui_pid: str
    :rtype: dict
    """
    reply_code = rh.REPLY_INTERNAL_SERVER_ERROR_500

    entry_point = ph.Pipeline.Rtsp.Ui.PIPELINE
    pipeline_jnode_set = helpers.get_pipeline_jnode_set2()
    if pipeline_jnode_set is None:
        eventlog.log_d("fail: internal-server-error")
        return rh.reply_fail(reply_code)

    # noinspection PyBroadException
    try:
        origin_uuid = str(uuid.uuid4())
        # create ui-jnode-set
        detectors_list = []
        ui_identity_set = _create_ui_jnode_set(ui_pid, origin_uuid, detectors_list)

        # create edge-jnode-set
        reply_code = _generator(ui_pid, ui_identity_set)
        if reply_code != rh.REPLY_SUCCESS_200:
            raise CreatorError

        # update pipeline.json
        visible_ui_type_set = map(lambda (ui_module, datasource): (ui_module.type, datasource),
                                  _UI_IDENTITY_TO_DATASOURCE_MAP)

        pipeline_jnode = ui_pipeline_set.add_pipeline(
            pipeline_jnode_set, ui_pid, origin_uuid,
            ph.PTYPE_LPR_RTSP, visible_ui_type_set,
            ui_detectors_list=detectors_list, internal=False)

        pipeline_ext_jnode = get_pipeline_ext_jnode(pipeline_jnode)

        # create time-jnode
        fs.create_jnode_from_default(_TIME_DEF_JNODE_FILEPATH, ph.UI_TIME_JNODE_FILEPATH)

        # update setupstore config
        append_ss_group(ui_pid, entry_point)

        # restart webserver with PCD
        helpers.restart_webserver_lazy()
        helpers.sighup_setupstore()

        return rh.reply_done(pipeline_ext_jnode)

    except CreatorError:
        pass

    except Exception:
        eventlog.log_x()

    delete_pipeline(ui_pid)
    return rh.reply_fail(reply_code)
