import os
import uuid
import datetime
import ujson

from edge import eventlog
from edge import path_helpers as ph
from edge.admin import reply_helpers as rh
from edge.admin import ui_types as ui
from edge import pipeline_helpers as oh
from edge.com import fs

import ui_pipeline_set
import edge_creator as creator
from _edge_setupstore import append_ss_group
from ui_pipeline import delete_pipeline, get_pipeline_ext_jnode

_TIME_DEF_JNODE_FILEPATH = '/opt/edge/share/setupstore/ui/time-jnode.json'

_UI_IDENTITY_TO_DATASOURCE_MAP = [
    (ph.Pipeline.Speed.Ui.EVENTSTORE_0, False),
    (ph.Pipeline.Speed.Ui.VIDEOSOURCE_0, True),
    (ph.Pipeline.Speed.Ui.LENS_0, True),
    (ph.Pipeline.Speed.Ui.GEOLOCATION_0, False),
    (ph.Pipeline.Speed.Ui.LPR_0, False),
    (ph.Pipeline.Speed.Ui.LANE_ENFORCEMENT_0, False),
    (ph.Pipeline.Speed.Ui.HTTP_STREAMING_0, False),
    (ph.Pipeline.Speed.Ui.RTSP_STREAMING_0, False),
    (ph.Pipeline.Speed.Ui.LOCATION_0, False)
]

_UI_DETECTORS_TO_INTERNAL_MAP = [
    (ph.Pipeline.Speed.Ui.TRANSIT_0, True)
]


def _create_ui_jnode_set(ui_pid, raw_body, origin_uuid, detectors_list):
    """
    :type ui_pid: str
    :type raw_body: str
    :type origin_uuid: str
    :type detectors_list: list
    :rtype: set of UiIdentity
    """
    args = ujson.loads(raw_body)
    serial = args['serial']

    # read def_jnode
    ui_identity_to_def_jnode_map = dict()
    for (ui_identity, _) in _UI_IDENTITY_TO_DATASOURCE_MAP:
        def_ui_jnode_filepath = ph.get_ui_def_jnode_filepath(ui_identity, raw_body)
        def_ui_jnode = fs.load_json2(def_ui_jnode_filepath)
        ui_identity_to_def_jnode_map[ui_identity] = def_ui_jnode

    # read detectors def jnode
    for (ui_identity, internal) in _UI_DETECTORS_TO_INTERNAL_MAP:
        def_ui_jnode_filepath = ph.get_ui_def_jnode_filepath(ui_identity, None)
        def_ui_jnode = fs.load_json2(def_ui_jnode_filepath)
        ui_identity_to_def_jnode_map[ui_identity] = def_ui_jnode
        detector_def_name = def_ui_jnode['rw'].get('args', {}).get('name', '')
        detectors_list.append((ui_identity, internal, detector_def_name))

    # edit
    ui_identity_to_def_jnode_map[ph.Pipeline.Speed.Ui.EVENTSTORE_0]['rw']['channel'] = int(ui_pid)
    ui_identity_to_def_jnode_map[ph.Pipeline.Speed.Ui.EVENTSTORE_0]['ro']['uuid'] = origin_uuid
    ui_identity_to_def_jnode_map[ph.Pipeline.Speed.Ui.EVENTSTORE_0]['ro']['ctime'] = datetime.datetime.now().isoformat()

    ui_identity_to_def_jnode_map[ph.Pipeline.Speed.Ui.VIDEOSOURCE_0]['rw']['serial'] = str(serial)

    # write
    ui_identity_set = set()
    for ui_identity, ui_jnode in ui_identity_to_def_jnode_map.iteritems():
        ui_identity_set.add(ui_identity)
        ui_jnode_dirpath = ph.get_ui_pipeline_jnode_dirpath(ui_pid, ui_identity)
        if not os.path.exists(ui_jnode_dirpath):
            fs.silent_mkdir_r(ui_jnode_dirpath)

        ui_jnode_filepath = ph.get_ui_pipeline_jnode_filepath(ui_pid, ui_identity.type, ui_identity.id)
        with open(ui_jnode_filepath, 'w') as json_file:
            ujson.dump(ui_jnode, json_file, indent=2)

        ui_schema_filepath = ph.get_ui_pipeline_schema_filepath(ui_pid, ui_identity.type, ui_identity.id)
        eventlog.log_d(ui_schema_filepath)
        fs.ln_schema(ui_schema_filepath, ui_jnode_filepath)

    return ui_identity_set


def _get_new_pid(pipeline_jnode_set):
    """
    :type pipeline_jnode_set: list of dict
    :rtype: str | None
    """
    used_ui_pid_set = set(map(lambda pipeline_jnode: int(pipeline_jnode['id']), pipeline_jnode_set))

    ui_pid = None
    for ui_pid_candidate in xrange(0, 32):
        if ui_pid_candidate not in used_ui_pid_set:
            ui_pid = str(ui_pid_candidate)
            eventlog.log_d("new-pipeline-id: %s" % ui_pid)
            break
    return ui_pid


def create_speed_pipeline(vendor, model, serial):
    """
    :type vendor: str
    :type model: str
    :type serial: str
    """
    # in this case mb better to get 3 stings and  create arg for "_create_ui_jnode_set": ??
    # vendor = ''
    # model = ''
    # serial = ''
    jnode = {
        "vendor": vendor,
        "model": model,
        "serial": serial
    }

    entry_point = ph.Pipeline.Speed.Ui.PIPELINE
    pipeline_jnode_set = oh.get_pipeline_jnode_set2()
    if pipeline_jnode_set is None:
        eventlog.log_d("fail: internal-server-error")
        return rh.reply_fail(rh.REPLY_INTERNAL_SERVER_ERROR_500)

    ui_pid = _get_new_pid(pipeline_jnode_set)
    if ui_pid is None:
        eventlog.log_d("fail: service-unavailable")
        return rh.reply_fail(rh.REPLY_SERVICE_UNAVAILABLE_503)

    # noinspection PyBroadException
    try:
        origin_uuid = str(uuid.uuid4())
        # create ui-jnode-set
        detectors_list = []
        ui_identity_set = _create_ui_jnode_set(ui_pid, ujson.dumps(jnode), origin_uuid, detectors_list)

        # create edge-jnode-set
        creator.generate(ui_pid, entry_point, None, ui_identity_set)

        # update pipeline.json
        visible_ui_type_set = map(lambda (module, datasource): (module.type, datasource),
                                  _UI_IDENTITY_TO_DATASOURCE_MAP)

        pipeline_jnode = ui_pipeline_set.add_pipeline(pipeline_jnode_set, ui_pid, origin_uuid,
                                                      entry_point.ptype, visible_ui_type_set,
                                                      ui_detectors_list=detectors_list, internal=True)
        pipeline_ext_jnode = get_pipeline_ext_jnode(pipeline_jnode)

        # create time-jnode
        fs.create_jnode_from_default(_TIME_DEF_JNODE_FILEPATH, ph.UI_TIME_JNODE_FILEPATH)

        # update setupstore config
        append_ss_group(ui_pid, entry_point)
        # restart webserver with PCD
        oh.restart_webserver_lazy()
        oh.sighup_setupstore()

        return rh.reply_done(pipeline_ext_jnode)

    except NotImplementedError:
        return 'fail: camera not supported'

    except ui.LockError:
        return rh.reply_fail(rh.REPLY_CONFLICT_409)

    except Exception:
        eventlog.log_x()

    delete_pipeline(ui_pid)
    eventlog.log_d('fail: internal-server-error')
    return rh.reply_fail(rh.REPLY_INTERNAL_SERVER_ERROR_500)
