import ujson

from edge import eventlog
import edge.pipeline_helpers as pipeline_helpers
from edge.path_helpers import UiIdentity

# _PIPELINE_SET_FILEPATH = '/etc/edge/setupstore/admin/actions/pipeline.json'
# _PIPELINE_DEFSET_FILEPATH = '/opt/edge/share/setupstore/ui/pipeline-jnode.json'


def validate_ui_pid(ui_pid):
    """
    :type ui_pid: str
    :rtype: bool
    """
    if not ui_pid:
        return False

    with open(pipeline_helpers.PIPELINE_SET_FILEPATH, 'r') as json_file:
        pipelines_jnode = ujson.load(json_file)
    if not pipelines_jnode:
        return False

    pipeline_jnode_set = filter(lambda item_jnode: item_jnode['id'] == ui_pid, pipelines_jnode)
    if len(pipeline_jnode_set) != 1:
        return False

    return True


def validate_ui_type(ui_pid, ui_type):
    """
    :type ui_pid: str
    :type ui_type: str
    :rtype: bool
    """
    if not (ui_pid and ui_type):
        return False

    with open(pipeline_helpers.PIPELINE_SET_FILEPATH, 'r') as json_file:
        pipelines_jnode = ujson.load(json_file)
    if not pipelines_jnode:
        return False

    pipeline_jnode_set = filter(lambda item_jnode: item_jnode['id'] == ui_pid, pipelines_jnode)
    if len(pipeline_jnode_set) != 1:
        return False

    # noinspection PyTypeChecker
    modules_jnode = pipeline_jnode_set[0]['modules']
    module_jnode_set = filter(lambda item_jnode: item_jnode['type'] == ui_type, modules_jnode)
    if len(module_jnode_set) != 1:
        return False

    return True


def validate_ui_id(ui_pid, ui_type, ui_id):
    """
    :type ui_pid: str
    :type ui_type: str
    :type ui_id: str
    :rtype: bool
    """
    if not (ui_pid and ui_type and ui_id):
        return False

    with open(pipeline_helpers.PIPELINE_SET_FILEPATH, 'r') as json_file:
        pipelines_jnode = ujson.load(json_file)
    if not pipelines_jnode:
        return False

    pipeline_jnode_set = filter(lambda item_jnode: item_jnode['id'] == ui_pid, pipelines_jnode)
    if len(pipeline_jnode_set) != 1:
        return False

    # noinspection PyTypeChecker
    modules_jnode = pipeline_jnode_set[0]['modules']
    module_jnode_set = filter(lambda item_jnode: item_jnode['type'] == ui_type and item_jnode['id'] == ui_id,
                              modules_jnode)
    if len(module_jnode_set) != 1:
        return False

    return True


def get_pipeline_type(ui_pid):
    """
    :type ui_pid: str
    :rtype: str | None
    """
    pipeline_jnode_set = pipeline_helpers.get_pipeline_jnode_set2()

    pipeline_jnode = filter(lambda item_jnode: item_jnode['id'] == ui_pid, pipeline_jnode_set)
    if len(pipeline_jnode) != 1:
        eventlog.log_d('fail: wrong pid')
        return None

    return pipeline_jnode[0]['type']


def add_pipeline(pipeline_set_jnode, ui_pid, uuid, ui_ptype, ui_modules_type_set,
                 ui_detectors_list=None, internal=False):
    """
    :type pipeline_set_jnode: list of dict
    :type ui_pid: str
    :type uuid: str
    :type ui_ptype: str
    :type ui_modules_type_set: list of (str, bool)
    :type ui_detectors_list: list of (UiIdentity, bool, str) | None
    :type internal: bool
    :rtype: list of dict
    """
    ui_modules_jnode = []
    for (ui_type, source) in ui_modules_type_set:
        ui_modules_jnode.append(
            {
                "datasource": source,
                "id": "0",
                "type": "%s" % ui_type,
                "title": "pipeline.%s" % ui_type
            })

    ui_detecors_jnode = []
    if ui_detectors_list:
        for (identity, internal_detector, detector_name) in ui_detectors_list:
            ui_detecors_jnode.append(
                {
                    "id": identity.id,
                    "type": "%s" % identity.type,
                    "internal": internal_detector,
                    "name": detector_name
                })

    pipeline_jnode = {
        "id": "%s" % ui_pid,
        "uuid": uuid,
        "internal": internal,
        "type": ui_ptype,
        "modules": ui_modules_jnode,
        "detectors": ui_detecors_jnode
    }

    pipeline_set_jnode.append(pipeline_jnode)

    pipeline_helpers.edit_db_origin(uuid, media=True, manual=True)

    with open(pipeline_helpers.PIPELINE_SET_FILEPATH, 'w') as json_file:
        ujson.dump(pipeline_set_jnode, json_file, indent=2)

    return pipeline_set_jnode


def remove_pipeline(ui_pid):
    """
    :type ui_pid: str
    :rtype: dict
    """
    origin_uuid = pipeline_helpers.get_origin_uuid(ui_pid)
    pipeline_helpers.edit_db_origin(origin_uuid, media=False, manual=False)

    with open(pipeline_helpers.PIPELINE_SET_FILEPATH, 'r') as json_file:
        pipeline_jnode_set = ujson.load(json_file)

    new_pipeline_set_jnode = filter(lambda pipeline_jnode: pipeline_jnode['id'] != ui_pid, pipeline_jnode_set)

    with open(pipeline_helpers.PIPELINE_SET_FILEPATH, 'w') as json_file:
        ujson.dump(new_pipeline_set_jnode, json_file, indent=2)

    return new_pipeline_set_jnode


def _get_new_detector_id(detector_jnode_set):
    """
    :type detector_jnode_set: list of dict
    :rtype: str | None
    """
    used_ui_id_set = set(map(lambda detector_jnode: int(detector_jnode['id']), detector_jnode_set))

    ui_id = None
    for ui_id_candidate in xrange(0, 32):
        if ui_id_candidate not in used_ui_id_set:
            ui_id = str(ui_id_candidate)
            eventlog.log_d("new-detector-id: %s" % ui_id)
            break
    return ui_id


def add_detector_to_pipeline(ui_pid, ui_type, name, datasource=False):
    """
    :type ui_pid: str
    :type ui_type: str
    :type name: str
    :type datasource: bool
    :rtype: dict
    """
    try:
        with open(pipeline_helpers.PIPELINE_SET_FILEPATH, 'r') as json_file:
            pipeline_set_jnode = ujson.load(json_file)
    except IOError as x:
        eventlog.log_d('fail: open pipeline.json (%s)' % x.message)
        return None

    pipeline_list_jnode = filter(lambda pipeline: pipeline['id'] == ui_pid, pipeline_set_jnode)

    if len(pipeline_list_jnode) > 1:
        eventlog.log_d('fail: more than one pipeline with same pid')
        return None

    new_id = _get_new_detector_id(pipeline_list_jnode[0]['detectors'])
    detector_jnode = {
        "id": "%s" % new_id,
        "type": ui_type,
        "name": name,
        "internal": False,
        "datasource": datasource
    }
    pipeline_list_jnode[0]['detectors'].append(detector_jnode)

    with open(pipeline_helpers.PIPELINE_SET_FILEPATH, 'w') as json_file:
        ujson.dump(pipeline_set_jnode, json_file, indent=2)

    for pipeline_jnode in pipeline_set_jnode:
        ptype = pipeline_jnode['type']
        pipeline_jnode['limits'] = {
            'detectors': pipeline_helpers.get_detector_types(ptype)
        }

    return new_id, pipeline_set_jnode


def delete_detector_from_pipeline(ui_pid, ui_type, ui_id, name):
    """
    :type ui_pid: str
    :type ui_type: str
    :type ui_id: str
    :type name: str
    :rtype: dict
    """
    with open(pipeline_helpers.PIPELINE_SET_FILEPATH, 'r') as json_file:
        pipeline_set_jnode = ujson.load(json_file)

    pipeline_list_jnode = filter(lambda pipeline: pipeline['id'] == ui_pid, pipeline_set_jnode)

    if len(pipeline_list_jnode) > 1:
        eventlog.log_d('fail: more than one pipeline with same pid')
        return None

    new_detectors_set = filter(lambda modules: modules['id'] != ui_id, pipeline_list_jnode[0]['detectors'])

    pipeline_list_jnode[0]['detectors'] = new_detectors_set

    with open(pipeline_helpers.PIPELINE_SET_FILEPATH, 'w') as json_file:
        ujson.dump(pipeline_set_jnode, json_file, indent=2)

    for pipeline_jnode in pipeline_set_jnode:
        ptype = pipeline_jnode['type']
        pipeline_jnode['limits'] = {
            'detectors': pipeline_helpers.get_detector_types(ptype)
        }

    return pipeline_set_jnode


def rename_detector(ui_pid, ui_type, ui_id, new_name):
    """
    :type ui_pid: str
    :type ui_type: str
    :type ui_id: str
    :type new_name: str
    :rtype: dict
    """
    with open(pipeline_helpers.PIPELINE_SET_FILEPATH, 'r') as json_file:
        pipeline_set_jnode = ujson.load(json_file)

    pipeline_list_jnode = filter(lambda pipeline: pipeline['id'] == ui_pid, pipeline_set_jnode)

    if len(pipeline_list_jnode) > 1:
        eventlog.log_d('fail: more than one pipeline with same pid')
        return None

    for detector in pipeline_list_jnode[0]['detectors']:
        if detector['id'] == ui_id and detector['type'] == ui_type:
            detector['name'] = new_name

    with open(pipeline_helpers.PIPELINE_SET_FILEPATH, 'w') as json_file:
        ujson.dump(pipeline_set_jnode, json_file, indent=2)

    for pipeline_jnode in pipeline_set_jnode:
        ptype = pipeline_jnode['type']
        pipeline_jnode['limits'] = {
            'detectors': pipeline_helpers.get_detector_types(ptype)
        }

    return pipeline_set_jnode
