
from edge import eventlog
from edge.fov import Fov
from edge.service import react, gpsd, tz
from edge.lane_detector import LaneDetector

from _kernel_args import KernelArgs
from .model.models import Track

from .services import stats
from .services import event
from .services import trainer
from .services import raas
from .services import origin
from .services import bs
from .services import plate_template
from .services import vs
from .services import tl
from .services import db
from .services import systest

from .detectors.types import IBaseDetector, IDetector
from .detectors.base.detector import create as mk_base_detector
from .detectors.transit.detector import create as mk_transit_detector
from .detectors.parking.detector import create as mk_parking_detector
from .detectors.wrong_way.detector import create as mk_wrong_way_detector
from .detectors.spot_speed.detector import create as mk_spot_speed_detector
from .detectors.red_light.detector import create as mk_red_ligth_detector


from .model.types import BaseEvent, Event


class BaseKernel:
    def __init__(self, stat_service, react_service, event_service, base_detector, detectors):
        """
        :type stat_service: stats.IStatService
        :type react_service: react.IReactService
        :type event_service: event.IEventService
        :type base_detector: IBaseDetector
        :type detectors: list[IDetector]
        """
        self._event_service = event_service

        self._detectors = detectors
        map(lambda x: x.set_handler(self._handle_event), detectors)

        self._base_detector = base_detector
        self._base_detector.set_handler(self._handle_base_event)

        self._react_service = react_service
        self._react_service.set_handler(self._handle_react)
        self._react_counters = stat_service.react

    def _handle_react(self, react_name, react_jnode):
        """
        :type react_name: str
        :type react_jnode: dict
        """
        try:
            track = Track.deserialize(react_jnode)
        except (KeyError, ValueError) as x:
            eventlog.log_e('fail: unpack {} <{}>'.format(react_name, x.message))
            return

        self._react_counters.inc_track(track)

        lr_track = track.lr_track
        if lr_track:
            init_quad = lr_track[0].plate_quad
            fini_quad = lr_track[-1].plate_quad
            self._react_counters.inc_plate_quad(init_quad, fini_quad)

        self._base_detector.analyze(track)

    def _handle_base_event(self, track, event_args):
        """
        :type track: Track
        :type event_args: BaseEvent
        """
        map(lambda x: x.analyze(track, event_args), self._detectors)

    def _handle_event(self, event_args):
        """
        :type event_args: Event
        """
        self._event_service.emit(event_args)


class Kernel(BaseKernel):
    def __init__(self, twisted_reactor, args):
        """
        :type args: KernelArgs
        """
        stat_service = stats.init_service(
            twisted_reactor, args.counters,
            stat_gpsd=bool(args.gpsd.endpoint))

        db_service = db.init_service(args.db)

        systest_service = systest.init_service(args.systest)

        lane_service = LaneDetector(
            args.lane_detector.id_to_lane_list,
            args.lane_detector.img_width,
            args.lane_detector.img_height)

        fov_service = Fov.create2(args.image_mapper) if args.image_mapper else None

        tz_service = tz.init_service(args.tz)
        bs_service = bs.init_service(args.bs, stat_service.bs)
        vs_service = vs.init_service(args.vs)
        raas_service = raas.init_service(twisted_reactor, args.raas)
        tl_service = tl.init_service(twisted_reactor, args.tl)
        trainer_service = trainer.init_service(twisted_reactor, args.trainer, tz_service)
        gpsd_service = gpsd.init_service(twisted_reactor, args.gpsd, stat_service.gpsd)
        origin_service = origin.init_service(args.origin, gpsd_service)
        event_service = event.init_service(stat_service.event, args.event_service, tz_service, bs_service,
                                           systest_service)
        template_service = plate_template.init_service()

        base_detector = mk_base_detector(
            tz_service, vs_service, raas_service, trainer_service,
            origin_service, lane_service, template_service, fov_service, args.base_detector)

        detectors = [mk_transit_detector(origin_service, tz_service, args.transit_args)]
        detectors.extend(map(
            lambda p_args: mk_parking_detector(tz_service, vs_service, origin_service, db_service, p_args),
            args.detectors.parking))

        detectors.extend(map(
            lambda ww_args: mk_wrong_way_detector(tz_service, vs_service, origin_service, lane_service, ww_args),
            args.detectors.wrong_way))

        detectors.extend(map(
            lambda ss_args: mk_spot_speed_detector(tz_service, vs_service, origin_service, lane_service, ss_args),
            args.detectors.spot_speed))

        detectors.extend(map(
            lambda rl_args: mk_red_ligth_detector(tz_service, vs_service, origin_service, tl_service, rl_args),
            args.detectors.red_light))

        react_service = react.init_service(twisted_reactor, [args.react_service.react_endpoint])

        def shutdown():
            react.fini_service(react_service)

            map(lambda detector: detector.destroy(), detectors)
            base_detector.destroy()

            event.fini_service(event_service)
            origin.fini_service(origin_service)
            gpsd.fini_service(gpsd_service)
            trainer.fini_service(trainer_service)
            raas.fini_service(raas_service)
            tl.fini_service(tl_service)
            bs.fini_service(bs_service)
            tz.fini_service(tz_service)
            vs.fini_service(vs_service)
            db.fini_service(db_service)
            systest.fini_service(systest_service)

        twisted_reactor.addSystemEventTrigger('before', 'shutdown', shutdown)

        BaseKernel.__init__(self, stat_service, react_service, event_service, base_detector, detectors)
        self._twisted_reactor = twisted_reactor
        self._counters = stats

    def run(self):
        self._twisted_reactor.run()
