from edge_bbox.types import RectF, RectI
from ...model.models import LrTrackItem
from ...services.types import TimeInterval


class Place:
    def __init__(self, r, x, y):
        """
        :type x: int
        :type y: int
        :type r: int
        """
        self._r = r
        self._min_x = x - r
        self._max_x = x + r
        self._min_y = y - r
        self._max_y = y + r

    def __contains__(self, point_xy):
        """
        :type point_xy: (int, int)
        :rtype: bool
        """
        same_x = self._min_x < point_xy[0] < self._max_x
        same_y = self._min_y < point_xy[1] < self._max_y
        return same_x and same_y

    def mk_plate_rect(self, scene_w, scene_h, plate_w, plate_h):
        """
        :type scene_w: int
        :type scene_h: int
        :type plate_w: int
        :type plate_h: int
        :rtype: RectF
        """
        wh_ratio = float(plate_w) / plate_h
        x = self._min_x + self._r
        y = self._min_y + self._r

        fl = (x - self._r * wh_ratio / 2.0) / float(scene_w)
        fr = (x + self._r * wh_ratio / 2.0) / float(scene_w)

        ft = (y - self._r / 2.0) / float(scene_h)
        fb = (y + self._r / 2.0) / float(scene_h)

        return RectF(fl, ft, fr, fb)

    @property
    def xy(self):
        """
        :rtype: (int, int)
        """
        return self._min_x + self._r, self._min_y + self._r

    @staticmethod
    def create(jnode):
        """
        :type jnode: list[int]
        :rtype: Place
        """
        if len(jnode) != 3:
            raise ValueError

        return Place(*jnode)

    def serialize(self):
        return [self._r, self._min_x + self._r, self._min_y + self._r]


class LogEntry:
    def __init__(self, place, ts_range, event=False):
        """
        :type place: Place
        :type ts_range: TimeInterval
        :type event: bool
        """
        self.place = place
        self.range = ts_range
        self.event = event

    @staticmethod
    def deserialize(jnode):
        """
        :type jnode: dict
        :rtype: LogEntry
        """
        place = Place.create(jnode['place'])
        ts_range = TimeInterval.create(jnode['range'])
        return LogEntry(place, ts_range, bool(jnode.get('event')))

    def serialize(self):
        """
        :rtype: dict
        """
        jnode = {
            'place': self.place.serialize(),
            'range': [self.range.min_ts, self.range.max_ts]
        }

        if self.event:
            jnode['event'] = True

        return jnode

    def set_event(self):
        self.event = True

    def clone(self):
        """
        :rtype: LogEntry
        """
        return LogEntry(self.place, TimeInterval(self.range.min_ts, self.range.max_ts), self.event)


class Log(list):
    def __init__(self, scene_wh, plate_wh=(0, 0), items=()):
        """
        :type scene_wh: (int, int)
        :type plate_wh: (int, int)
        :type items: list[LogEntry]
        """
        self.scene_wh = scene_wh
        self.plate_w, self.plate_h = plate_wh
        list.__init__(self, items)

    @staticmethod
    def deserialize(jnode):
        """
        :type jnode: dict
        :rtype: Log
        """
        scene_wh_jnode = jnode['scene-wh']
        plate_wh_jnode = jnode['plate-wh']
        items = map(lambda item_jnode: LogEntry.deserialize(item_jnode),
                    jnode['timeline'])

        return Log(scene_wh_jnode, plate_wh_jnode, items)

    def serialize(self):
        """
        :rtype: dict | null  # TODO: del in db, if null
        """
        return {
            'scene-wh': self.scene_wh,
            'plate-wh': [self.plate_w, self.plate_h],
            'timeline': map(lambda item: item.serialize(), self)
        }

    def set_event(self, event_ts):
        """
        :type event_ts: int
        """
        for log_entry in self:  # type: LogEntry
            if log_entry.range.min_ts <= event_ts <= log_entry.range.max_ts:
                log_entry.set_event()
                break

    def prolong(self, lr_items, range_tout, lost_tout):
        """
        :type lr_items: list[LrTrackItem]
        :type range_tout: int
        :type lost_tout: int
        """
        init_item = lr_items[0]
        init_rect = RectI(*init_item.plate_quad.get_abs_bbox())

        # cleanup
        if len(self):
            if (init_item.ts - self[-1].range.max_ts) > lost_tout:
                del self[:]  # self.clear()
                #print 'lost-tout'

        #
        #     elif init_rect.center not in self[-1].place:
        #         # place changed
        #         print 'place changed', \
        #             self[-1].place.x, \
        #             self[-1].place.y, \
        #             self[-1].place.r, \
        #             init_rect.center
        #
        #         del self[:]  # self.clear()

        # prolong empty
        if len(self):
            entry = self[-1]

        else:
            entry = LogEntry(
                Place(init_rect.ih, *init_rect.center),
                TimeInterval(init_item.ts, init_item.ts))
            self.append(entry)

        for next_item in lr_items:  # type: LrTrackItem
            next_rect = RectI(*next_item.plate_quad.get_abs_bbox())

            # upscale plate-wh
            if next_rect.iw > self.plate_w and next_rect.ih > self.plate_h:
                self.plate_w = next_rect.iw
                self.plate_h = next_rect.ih

            if next_rect.center in entry.place and (next_item.ts - entry.range.max_ts) < range_tout:
                entry.range.max_ts = next_item.ts

            else:
                entry = LogEntry(
                    Place(next_rect.ih, *next_rect.center),
                    TimeInterval(next_item.ts, next_item.ts))
                self.append(entry)

        #print 'prolong', self.serialize()

    def mk_trend(self, lost_tout):
        """
        :type lost_tout: int
        :rtype: Log
        """

        trend_entry = self[0].clone()  # type: LogEntry
        trend_items = [trend_entry]  # type: list[LogEntry]

        for log_entry in self:  # type: LogEntry

            if log_entry.place.xy in trend_entry.place:
                if (log_entry.range.min_ts - trend_items[-1].range.max_ts) < lost_tout:
                    trend_entry.range.max_ts = log_entry.range.max_ts
                    if log_entry.event:
                        trend_entry.set_event()

                    continue

            trend_entry = log_entry.clone()
            trend_items.append(trend_entry)

        return Log(self.scene_wh, (self.plate_w, self.plate_h), trend_items)
