from edge import eventlog
from edge_bbox.types import mk_plate_bbox, RectF
from edge.lane_detector import LaneDetector
from edge.image.image import ImEnv, ImEngine, ImSizeFitter
from edge.image.types import ImFooterVars
from edge.service.tz import ITzService
from edge.model.types import Dir

from .types import DetectorArgs, RuleArgs, SpotSpeedEvidence
from ..base.best import check_quad_dist_from_img_borders
from ..types import IDetector
from ..com.schedule import Schedule
from ..com.types import VideoArgs
from ...model.models import Track, TrTrackItem
from ...model.types import Event, Position, ViolationWaypoint, BaseEvent
from ...services.origin import IOriginService
from ...services.vs import IVsService, VsImageInfo, VsBufferTooSmall, VsError


def get_median_speed(tr_items_list):
    """
    :type tr_items_list: list of (TrTrackItem)
    :rtype: int | None
    """
    # return 80
    ra_trace = filter(lambda tr_item: tr_item.ra_item, tr_items_list)

    speed_list = map(lambda ra_scene_item: ra_scene_item.ra_item.speed, ra_trace)

    speed_list_len = len(speed_list)
    if speed_list_len > 0:
        speed_list.sort()
        index = speed_list_len / 2
        return int(speed_list[index])
    #
    # return None


class _SceneMsRef(VsImageInfo):
    def __init__(self, tr_item):
        """
        :type tr_item: TrTrackItem
        """
        VsImageInfo.__init__(self, tr_item.ts)
        self.tr_item = tr_item


class RuleServices:
    def __init__(self, tz_service, vs_service, origin_service):
        """
        :type tz_service: ITzService
        :type vs_service: IVsService
        :type origin_service: IOriginService
        """
        self.tz = tz_service
        self.vs = vs_service
        self.origin = origin_service


class Rule:
    def __init__(self, services, rule_args, detector_name, init_im_env, fini_im_env, video_args):
        """
        :type services: RuleServices
        :type rule_args: RuleArgs
        :type detector_name: str
        :type init_im_env: ImEnv
        :type fini_im_env: ImEnv
        :type video_args: VideoArgs | None
        :type
        """
        self._detector_name = detector_name

        self._init_im_env = init_im_env
        self._fini_im_env = fini_im_env

        self._services = services
        self._scheduler = Schedule(rule_args.schedule, services.tz)
        self._args = rule_args

        self._video = video_args
        if video_args:
            self._im_fitter = ImSizeFitter((video_args.max_w, video_args.max_h))

    def check(self, track, base, lane_to_tr_item_list):
        """
        :type track: Track
        :type base: BaseEvent
        :type lane_to_tr_item_list: list of (str, TrTrackItem)
        :rtype: SpotSpeedEvidence | None
        """
        cts = track.lr_track.min_ts
        lts = track.lr_track.max_ts

        if not(self._scheduler.check2(lts) or self._scheduler.check2(cts)):
            return None

        tr_items_list = []
        for lane, tr_item in lane_to_tr_item_list:
            if lane in self._args.limits.lanes or len(self._args.limits.lanes) == 0:
                # we use only points where we have both lr and ra data
                if tr_item.lr_item and tr_item.ra_item:
                    lr_scene_img = track.lr_track.best_image
                    # filter lr track near to img edge
                    if check_quad_dist_from_img_borders(tr_item.lr_item.plate_quad,
                                                        lr_scene_img.width, lr_scene_img.height):
                        tr_items_list.append(tr_item)

        if len(tr_items_list) < 2:
            return None

        speed = get_median_speed(tr_items_list)

        if speed is None or speed < (self._args.limits.speed_limit + self._args.limits.over_speed):
            return None

        # overspeed
        scene_ms_refs = map(_SceneMsRef, tr_items_list)
        init_scene_ms_refs = scene_ms_refs[0:5]
        # init_scene_ms_refs.sort(key=lambda ref: ref.ts)

        fini_scene_ms_refs = scene_ms_refs[:-6:-1]
        # fini_scene_ms_refs.sort(key=lambda ref: ref.ts, reverse=True)

        try:
            init_winner_img = self._services.vs.export_image_rgba(self._init_im_env.scene_rgba, init_scene_ms_refs)
        except VsBufferTooSmall as ex:
            self._init_im_env.scene_rgba.reset(14, ex.min_w, ex.min_h, self._init_im_env.footer_h)
            init_winner_img = self._services.vs.export_image_rgba(self._init_im_env.scene_rgba, init_scene_ms_refs)

        try:
            fini_winner_img = self._services.vs.export_image_rgba(self._fini_im_env.scene_rgba, fini_scene_ms_refs)
        except VsBufferTooSmall as ex:
            self._fini_im_env.scene_rgba.reset(14, ex.min_w, ex.min_h, self._fini_im_env.footer_h)
            fini_winner_img = self._services.vs.export_image_rgba(self._fini_im_env.scene_rgba, fini_scene_ms_refs)

        if None in [init_winner_img, fini_winner_img]:
            eventlog.log_d('fail: get init/fini image from vs')
            return None

        return self._mk_evidence(base, track, speed, init_winner_img.tr_item, fini_winner_img.tr_item)

    def _mk_evidence(self, base_event, track, speed, init_tr_item, fini_tr_item):
        """
        :type init_tr_item: TrTrackItem
        :type fini_tr_item: TrTrackItem
        :type speed: int
        :type track Track
        :type base_event: BaseEvent
        :rtype: SpotSpeedEvidence | None
        """
        plate_text = base_event.get_plate_text()

        # workaround
        from ...model.types import BestWaypoint, Vehicle
        best_waypoint = base_event.measurements[0].waypoints[2]  # type: BestWaypoint
        best_vehicle = best_waypoint.vehicle  # type: Vehicle
        camera_speed = best_vehicle.camera_speed

        def mk_waypoint(services, im_env, tr_item):
            """
            :type services: RuleServices
            :type im_env: ImEnv
            :type tr_item: TrTrackItem
            :rtype: ViolationWaypoint
            """
            plate_rect = RectF(*tr_item.lr_item.plate_quad.get_rel_bbox(track.lr_track.best_image.width,
                                                                        track.lr_track.best_image.height))
            plate_bbox = mk_plate_bbox(plate_rect, *im_env.scene_rgba.wh)

            waypoint_origin = services.origin.get(im_env.scene_rgba.ts)
            waypoint_ts = im_env.scene_rgba.ts
            waypoint_position = Position(tr_item.ra_item.raw_f, tr_item.ra_item.raw_l)

            footer_vars = ImFooterVars(
                im_env.footer_def_vars, waypoint_ts,
                waypoint_origin.location,
                waypoint_origin.geolocation,
                plate_text,
                spot_speed=speed,
                tz_service=services.tz)

            im_blobs = im_env.engine.mk_blobs(
                im_env.scene_rgba, plate_bbox, footer_vars, track.lr_track, tr_item.lr_item.plate_quad,
                vehicle=True, scene=True, plate=True)

            if len(im_blobs) != 3:
                eventlog.log_d('fail: get all blobs for evidence')
                raise RuntimeError

            wp_direction = track.tr_track.get_item_direction(tr_item)
            waypoint = ViolationWaypoint(waypoint_ts, waypoint_origin.location,
                                         direction=wp_direction,
                                         position=waypoint_position,
                                         plate_quads=[tr_item.lr_item.plate_quad])

            waypoint.add_images(im_blobs)
            return waypoint

        try:
            init_wp = mk_waypoint(self._services, self._init_im_env, init_tr_item)
            fini_wp = mk_waypoint(self._services, self._fini_im_env, fini_tr_item)
        except RuntimeError:
            return None

        event_ts = self._fini_im_env.scene_rgba.ts
        direction = Dir(Dir.UNKNOWN)
        init_position = Position(init_tr_item.ra_item.raw_f, init_tr_item.ra_item.raw_l)
        fini_position = Position(fini_tr_item.ra_item.raw_f, fini_tr_item.ra_item.raw_l)
        if init_position.forward > fini_position.forward:
            direction = Dir(Dir.APPROACHING)
        elif init_position.forward < fini_position.forward:
            direction = Dir(Dir.RECEDING)

        evidence = SpotSpeedEvidence(self._services.origin.get(event_ts), event_ts, speed, camera_speed,
                                     self._args, self._detector_name, direction)
        evidence.add_waypoints([
            init_wp,
            fini_wp
        ])

        if self._video:
            scaled_w, scaled_h = self._im_fitter.scaledown(self._init_im_env.scene_rgba.wh)
            try:
                ss_video = self._services.vs.export_video(
                    self._init_im_env.scene_rgba.ts / 1000 - self._video.pre_record,
                    self._fini_im_env.scene_rgba.ts / 1000 + self._video.post_record,
                    scaled_w, scaled_h, self._video.quality)
                evidence.add_videos([ss_video])
            except VsError:
                eventlog.log_d('fail: save video')

        return evidence


def create(*args, **kwargs):
    """
    :rtype: IDetector
    """
    return _Detector(*args, **kwargs)


class _Detector(IDetector):
    def __init__(self, tz_service, vs_service, origin_service, lane_detector, args):
        """
        :type tz_service: ITzService
        :type vs_service: IVsService
        :type origin_service: IOriginService
        :type lane_detector: LaneDetector
        :type args: DetectorArgs
        """
        self._identity = args.identity
        self._lane_detector = lane_detector

        self._handler = None

        self._init_im_env = ImEnv(
            ImEngine(args.im_engine),
            args.im_engine.scene.footer.height,
            args.def_footer_mapping)

        self._fini_im_env = ImEnv(
            self._init_im_env.engine,
            self._init_im_env.footer_h,
            self._init_im_env.footer_def_vars)

        self._rules = []
        rule_services = RuleServices(tz_service, vs_service, origin_service)
        for rule in args.rules:
            if rule.enabled:
                self._rules.append(Rule(rule_services, rule,
                                        args.name,
                                        self._init_im_env,
                                        self._fini_im_env,
                                        args.video))

        IDetector.__init__(self)

    def analyze(self, track, base):
        """
        :type track: Track
        :type base: BaseEvent
        """
        if None in [track.lr_track, track.ra_track]:
            return
        # if len(track.ra_track) == 0:
        #     return

        def get_parts():
            """
            :rtype: list of (str, TrTrackItem)
            """
            lane_to_tr_item = []

            for tr_item in track.tr_track:  # type: TrTrackItem
                if tr_item.lr_item:
                    lr_scene_img = track.lr_track.best_image
                    lr_rect = RectF(*tr_item.lr_item.plate_quad.get_rel_bbox(lr_scene_img.width, lr_scene_img.height))
                    lane_set = self._lane_detector.get_lane2(lr_rect.fl, lr_rect.ft, lr_rect.fr, lr_rect.fb)
                    if len(lane_set) != 1:
                        lane_to_tr_item.append(('no_lane', tr_item))
                        continue

                    lane_name = lane_set.pop()
                    lane_to_tr_item.append((lane_name, tr_item))

            return lane_to_tr_item

        lane_to_tr_item_map = get_parts()

        for rule in self._rules:
            evidence = rule.check(track, base, lane_to_tr_item_map)

            if evidence:
                self._handler(Event(self._identity, base, evidence))
                break

            else:
                pass

    def set_handler(self, event_handler):
        """
        :type event_handler: (Event) -> None
        """
        if self._handler:
            raise NotImplementedError

        self._handler = event_handler
