from edge import eventlog
from edge_bbox.types import mk_plate_bbox, QuadI, RectF
from edge.lane_detector import LaneDetector
import math

from edge.image.image import ImEnv, ImEngine, ImSizeFitter
from edge.image.types import ImFooterVars
from edge.service.tz import ITzService
from edge.model.types import Dir

from .types import DetectorArgs, WrongWayEvidence, RuleArgs
from ..com.scene import SceneMask
from ..com.types import VideoArgs
from ..base.best import check_quad_dist_from_img_borders
from ..types import IDetector
from ...detectors.com.schedule import Schedule
from ...model.types import BaseEvent, Event, ViolationWaypoint
from ...model.models import Track, TrTrackItem
from ...services.origin import IOriginService
from ...services.vs import IVsService, VsImageInfo, VsBufferTooSmall, VsError


def points_far_enought(quad1, quad2):
    """
    :type quad1: QuadI
    :type quad2: QuadI
    :rtype: bool
    """
    _, plate_h = quad1.wh
    max_distance = plate_h * 0.7  # % of plate height

    rect1_c_x, rect1_c_y = quad1.center
    rect2_c_x, rect2_c_y = quad2.center

    if math.fabs(rect1_c_x - rect2_c_x) < max_distance and math.fabs(rect1_c_y - rect2_c_y) < max_distance:
        # points are in same rect
        return False
    else:
        # points are far enought
        return True


def filter_tr_items(tr_items, lr_scene_w, lr_scene_h):
    """
    :type tr_items: list of TrTrackItem
    :type lr_scene_w: int
    :type lr_scene_h: int
    :rtype: list of TrTrackItem
    """
    result = []
    # todo: lr_item can be null ?
    init_point = tr_items[0]  # type: TrTrackItem
    result.append(init_point)

    for index in range(1, len(tr_items), 1):
        fini_point = tr_items[index]  # type: TrTrackItem
        if fini_point.lr_item is None:
            continue

        if not check_quad_dist_from_img_borders(fini_point.lr_item.plate_quad, lr_scene_w, lr_scene_h):
            continue

        if points_far_enought(init_point.lr_item.plate_quad, fini_point.lr_item.plate_quad):
            result.append(fini_point)
            init_point = fini_point

    return result


def check_direction(quad1, quad2):
    """
    :type quad1: QuadI
    :type quad2: QuadI
    :rtype: Dir
    """
    _, plate_h = quad1.wh
    max_dist = plate_h * 0.4  # % of plate h

    rect1_c_x, rect1_c_y = quad1.center

    rect2_c_x, rect2_c_y = quad2.center

    if max_dist < rect2_c_y - rect1_c_y:
        direction = Dir(Dir.APPROACHING)

    elif rect2_c_y - rect1_c_y < - max_dist:
        direction = Dir(Dir.RECEDING)

    else:
        direction = Dir(Dir.UNKNOWN)

    return direction


class _SceneMsRef(VsImageInfo):
    def __init__(self, tr_item):
        """
        :type tr_item: TrTrackItem
        """
        VsImageInfo.__init__(self, tr_item.lr_item.ts)
        self.tr_item = tr_item


class RuleServices:
    def __init__(self, tz_service, vs_service, origin_service):
        """
        :type tz_service: ITzService
        :type vs_service: IVsService
        :type origin_service: IOriginService
        """
        self.tz = tz_service
        self.vs = vs_service
        self.origin = origin_service


class Rule:
    def __init__(self, services, rule_args, detector_name, init_im_env, fini_im_env, video_args):
        """
        :type services: RuleServices
        :type rule_args: RuleArgs
        :type detector_name: str
        :type init_im_env: ImEnv
        :type fini_im_env: ImEnv
        :type video_args: VideoArgs | None
        :type
        """
        self._detector_name = detector_name

        self._init_im_env = init_im_env
        self._fini_im_env = fini_im_env

        self._services = services
        self._scheduler = Schedule(rule_args.schedule, self._services.tz)
        self._args = rule_args

        self._video = video_args
        if video_args:
            self._im_fitter = ImSizeFitter((video_args.max_w, video_args.max_h))

    def check(self, track, base, lane_to_tr_item):
        """
        :type track: Track
        :type base: BaseEvent
        :type lane_to_tr_item: list of (str, TrTrackItem)
        :rtype: Deferred | None
        """
        # print 'rule %s' % self._args.name
        cts = track.lr_track.min_ts
        lts = track.lr_track.max_ts

        if not(self._scheduler.check2(lts) or self._scheduler.check2(cts)):
            return None

        lane_match_list = []
        for lane, tr_item in lane_to_tr_item:
            if lane in self._args.limits.lanes or len(self._args.limits.lanes) == 0:
                lane_match_list.append(tr_item)

        if len(lane_match_list) < 3:
            return None

        # lane_match_list.sort(key=lambda ref: ref.ts, reverse=False)

        violation_tr_items = []
        if self._args.limits.direction.dir == Dir.NOWAY:
            # violation on all items from lane_match_list
            violation_tr_items = lane_match_list
        else:
            # check real direction
            dist = 2
            for index in range(dist, len(lane_match_list), 1):
                init_point = lane_match_list[index - dist]  # type: TrTrackItem
                fini_point = lane_match_list[index]  # type: TrTrackItem

                direction = track.tr_track.get_item_direction(fini_point, max_lookup_depth=2)
                # direction = check_direction(init_point.lr_item.plate_quad, fini_point.lr_item.plate_quad)
                #direction = check_direction(track.lr_track[0].plate_quad,
                                            #track.lr_track[-1].plate_quad)
                #print direction.__str__(), fini_point.lr_item.plate_quad.center[1], self._services.tz.get_datetime_isostr(fini_point.ts/ 1e6)
                if direction in [Dir(Dir.APPROACHING), Dir(Dir.RECEDING)] and direction == self._args.limits.direction:
                    violation_tr_items.append(lane_match_list[index])

        if len(violation_tr_items) < 2:
            return

        scene_ms_refs = map(_SceneMsRef, violation_tr_items)
        init_scene_ms_refs = scene_ms_refs[0:5]
        # init_scene_ms_refs.sort(key=lambda ref: ref.ts)

        fini_scene_ms_refs = scene_ms_refs[:-6:-1]
        # fini_scene_ms_refs.sort(key=lambda ref: ref.ts, reverse=True)

        try:
            init_winner_img = self._services.vs.export_image_rgba(self._init_im_env.scene_rgba, init_scene_ms_refs)
        except VsBufferTooSmall as ex:
            self._init_im_env.scene_rgba.reset(14, ex.min_w, ex.min_h, self._init_im_env.footer_h)
            init_winner_img = self._services.vs.export_image_rgba(self._init_im_env.scene_rgba, init_scene_ms_refs)

        try:
            fini_winner_img = self._services.vs.export_image_rgba(self._fini_im_env.scene_rgba, fini_scene_ms_refs)
        except VsBufferTooSmall as ex:
            self._fini_im_env.scene_rgba.reset(14, ex.min_w, ex.min_h, self._fini_im_env.footer_h)
            fini_winner_img = self._services.vs.export_image_rgba(self._fini_im_env.scene_rgba, fini_scene_ms_refs)

        if None in [init_winner_img, fini_winner_img]:
            eventlog.log_d('fail: get init/fini image from vs')
            return None

        return self._mk_evidence(base, track, init_winner_img.tr_item, fini_winner_img.tr_item)

    def _mk_evidence(self, base_event, track, init_tr_item, fini_tr_item):
        """
        :type init_tr_item: TrTrackItem
        :type fini_tr_item: TrTrackItem
        :type track Track
        :type base_event: BaseEvent
        :return:
        """
        plate_text = base_event.get_plate_text()

        def mk_waypoint(services, im_env, tr_item):
            """
            :type services: RuleServices
            :type im_env: ImEnv
            :type tr_item: TrTrackItem

            :rtype: ViolationWaypoint

            :raise RuntimeError:
            """
            lr_scene_img = track.lr_track.best_image
            plate_rect_f = RectF(*tr_item.lr_item.plate_quad.get_rel_bbox(lr_scene_img.width, lr_scene_img.height))
            plate_bbox = mk_plate_bbox(plate_rect_f, *im_env.scene_rgba.wh)

            waypoint_origin = services.origin.get(im_env.scene_rgba.ts)
            waypoint_ts = im_env.scene_rgba.ts

            footer_vars = ImFooterVars(
                im_env.footer_def_vars, waypoint_ts,
                waypoint_origin.location,
                waypoint_origin.geolocation,
                plate_text,
                spot_speed=tr_item.ra_item.speed if tr_item.ra_item else None,
                tz_service=services.tz)

            im_blobs = im_env.engine.mk_blobs(
                im_env.scene_rgba, plate_bbox, footer_vars, track.lr_track, tr_item.lr_item.plate_quad,
                vehicle=False, scene=True, plate=True)

            if len(im_blobs) != 2:
                eventlog.log_d('fail: get all blobs for evidence')
                raise RuntimeError

            wp_direction = track.tr_track.get_item_direction(tr_item)

            waypoint = ViolationWaypoint(
                waypoint_ts, waypoint_origin.location,
                position=None,
                direction=wp_direction,
                plate_quads=[tr_item.lr_item.plate_quad])

            waypoint.add_images(im_blobs)
            return waypoint

        try:
            init_wp = mk_waypoint(self._services, self._init_im_env, init_tr_item)
            fini_wp = mk_waypoint(self._services, self._fini_im_env, fini_tr_item)
        except RuntimeError:
            return None

        init_ts = self._init_im_env.scene_rgba.ts
        fini_ts = self._fini_im_env.scene_rgba.ts
        event_ts = fini_ts
        duration = int((fini_ts - init_ts) / 1000000)
        evidence = WrongWayEvidence(self._services.origin.get(event_ts), event_ts, duration, self._args,
                                    self._detector_name)
        evidence.add_waypoints([
            init_wp,
            fini_wp
        ])

        if self._video:
            scaled_w, scaled_h = self._im_fitter.scaledown(self._init_im_env.scene_rgba.wh)
            try:
                ww_video = self._services.vs.export_video(
                    track.lr_track.min_ts / 1000 - self._video.pre_record,
                    track.lr_track.max_ts / 1000 + self._video.post_record,
                    scaled_w, scaled_h, self._video.quality)
                evidence.add_videos([ww_video])
            except VsError:
                eventlog.log_d('fail: save video')

        return evidence


def create(*args, **kwargs):
    """
    :rtype: IDetector
    """
    return _Detector(*args, **kwargs)


class _Detector(IDetector):
    def __init__(self, tz_service, vs_service, origin_service, lane_service, args):
        """
        :type vs_service: IVsService
        :type tz_service: ITzService
        :type lane_service: LaneDetector
        :type args: DetectorArgs
        """
        self._ruleNameToSheduler = {}
        self._lane_detector = lane_service
        self._scene = args.scene

        self._init_im_env = ImEnv(
            ImEngine(args.im_engine),
            args.im_engine.scene.footer.height,
            args.def_footer_mapping)

        self._fini_im_env = ImEnv(
            self._init_im_env.engine,
            self._init_im_env.footer_h,
            self._init_im_env.footer_def_vars)

        self._rules = []
        rule_services = RuleServices(tz_service, vs_service, origin_service)
        for rule in args.rules:
            if rule.enabled:
                self._rules.append(Rule(rule_services, rule,
                                        args.name,
                                        self._init_im_env,
                                        self._fini_im_env,
                                        args.video))

        self._identity = args.identity
        self._handler = None
        IDetector.__init__(self)

    def set_handler(self, event_handler):
        """
        :type event_handler: (Event) -> None
        """
        if self._handler:
            raise NotImplementedError

        self._handler = event_handler

    def analyze(self, track, base):
        """
        :type track: Track
        :type base: BaseEvent
        """
        if track.lr_track is None:
            return

        sc_detector = SceneMask(self._scene)

        def get_parts():
            """
            :rtype: list of (str, TrTrackItem)
            """
            lane_to_tritem = []

            lr_scene_img = track.lr_track.best_image
            # for tr_item in filter_tr_items(track.tr_track, lr_scene_img.width, lr_scene_img.height):
            for tr_item in track.tr_track:
                if tr_item.lr_item:
                    if not sc_detector.check_quad(tr_item.lr_item.plate_quad):
                        continue

                    # TODO:
                    lr_scene_img = track.lr_track.best_image
                    lr_rect = RectF(*tr_item.lr_item.plate_quad.get_rel_bbox(lr_scene_img.width, lr_scene_img.height))

                    lane_set = self._lane_detector.get_lane2(lr_rect.fl, lr_rect.ft, lr_rect.fr, lr_rect.fb)
                    if len(lane_set) != 1:
                        lane_to_tritem.append(('no_lane', tr_item))
                        continue

                    lane_name = lane_set.pop()
                    lane_to_tritem.append((lane_name, tr_item))

            return lane_to_tritem

        lane_to_tr_item = get_parts()

        for rule in self._rules:
            evidence = rule.check(track, base, lane_to_tr_item)
            if evidence:
                self._handler(Event(self._identity, base, evidence))
                # print 'violation on rule %s' % rule._args.name
                break
