from uuid import uuid4
import bisect

from edge.types import ImageRef
from edge_bbox.types import RectF, QuadI
from edge.model.track import ITrack

from edge.model.types import Dir


class RaTrackItem:
    def __init__(self, ts, x, y, speed, raw_f, raw_l):
        """
        :type ts: int
        :type x: float
        :type y: float
        :type speed: int
        """
        self.ts = ts
        self.x = x
        self.y = y
        self.speed = speed
        self.raw_f = raw_f
        self.raw_l = raw_l

    @staticmethod
    def deserialize(ra_track_item_jnode):
        """
        :type ra_track_item_jnode: dict
        :rtype: RaTrackItem
        """
        ts = ra_track_item_jnode['Ts']
        x, y = ra_track_item_jnode['Xy']
        speed_mps = ra_track_item_jnode['S']
        raw_f = ra_track_item_jnode.get('F', 0)
        raw_l = ra_track_item_jnode.get('L', 0)

        ra_track_item = RaTrackItem(ts, x, y, speed_mps, raw_f, raw_l)
        return ra_track_item


class RaTrack(list):
    def __init__(self, length, trace):
        """
        :type length: int
        :type trace: list[RaTrackItem]
        """
        list.__init__(self, trace)
        self.length = length

    def get_median_speed_kmph(self):
        """
        :rtype: int | None
        """
        speed_list = map(lambda item: item.speed, self)

        speed_list_len = len(speed_list)
        if speed_list_len > 0:
            speed_list.sort()
            index = speed_list_len / 2
            return int(speed_list[index])

    @staticmethod
    def deserialize(jnode):
        """
        :param jnode: TrTrack#/Ra
        :type jnode: dict
        :rtype: RaTrack
        """
        length = jnode['Length']
        trace = map(lambda item_jnode: RaTrackItem.deserialize(item_jnode), jnode['Trace'])
        return RaTrack(length, trace)


class Plate:
    def __init__(self, validity, text, template, quad=None):
        """
        :type validity: int
        :type text: str
        :type template: int
        :type quad: LrQuad | None
        """
        self.validity = validity
        self.text = text
        self.template = template
        self.quad = quad

    @staticmethod
    def deserialize_lr(plate_jnode):
        """
        :type plate_jnode: dict
        :rtype: Plate
        """
        validity = plate_jnode['V']
        text = plate_jnode['P']
        template = plate_jnode['T']
        quad = QuadI.unpack(plate_jnode['Q'])

        return Plate(validity, text, template, quad)

    @staticmethod
    def deserialize_raas(jnode):
        """
        :param jnode: RaasScene#/hypos/:index
        :type jnode: dict
        :rtype: Plate
        """
        validity = jnode['weight']
        text = jnode['number']
        country = int(jnode['country']) << 16
        template = country | int(jnode['template'])

        return Plate(validity, text, template)


class RaasSceneItem:
    def __init__(self, variants):
        """
        :type variants: list[RaasPlate]
        """
        self.variants = variants

    @staticmethod
    def deserialize(jnode):
        """
        :param jnode: RaasScene#/hypos/:index
        :type jnode: dict
        :rtype: RaasSceneItem
        """
        plate = Plate.deserialize_raas(jnode)
        return RaasSceneItem([plate])


class RaasScene:
    def __init__(self, items):
        """
        :type items: list[RaasSceneItem]
        """
        self.items = items

    def try_get_first_scene_item_plate(self):
        """
        :rtype: Plate | None
        """
        for item in self.items:
            max_validity_variant = max(item.variants, key=lambda entry: entry.validity)
            return max_validity_variant

    @staticmethod
    def deserialize(jnode):
        """
        :param jnode: RaasScene#/
        :type jnode: dict
        :rtype: RaasScene

        :raise ValueError:
        """
        if 0 != jnode['result'] and 1 != jnode['version']:
            raise ValueError

        items = map(
            lambda raas_scene_item_jnode: RaasSceneItem.deserialize(raas_scene_item_jnode),
            jnode['hypos'])

        return RaasScene(items)


class LrTrackItem:
    def __init__(self, ts, plate_weight, plate_quad):
        """
        :type ts: int
        :type plate_weight: int
        :type plate_quad: QuadI
        """
        self.ts = ts
        self.plate_weight = plate_weight
        self.plate_quad = plate_quad

    @staticmethod
    def deserialize(jnode):
        """
        :param jnode: TrTrack#/Lr/PlateTrace/:index
        :type jnode: dict
        :rtype: LrTrackItem
        """
        ts = jnode['Ts']
        weight = jnode['W']
        quad = QuadI.unpack(jnode['Q'])

        return LrTrackItem(ts, weight, quad)


class JsonImRef(ImageRef):
    def __init__(self, *args, **kwargs):
        ImageRef.__init__(self, *args, **kwargs)

    @staticmethod
    def deserialize(jnode):
        """
        :type jnode: dict
        :rtype: JsonImRef
        """
        blob_ts = jnode['Ts']
        blob_url = jnode.get('Url')
        if blob_url is None or len(blob_url) == 0:
            blob_url = None

        blob_format = jnode['Codec']
        blob_w, blob_h = jnode['Wh']

        return JsonImRef(blob_ts, blob_url, blob_format, int(blob_w), int(blob_h))


class LrTrack(ITrack, list):
    def __init__(self, plate, plate_trace, image_dir, image_ref):
        """
        :type image_dir: int
        :type plate: Plate
        :type plate_trace: list[LrTrackItem]
        :type image_ref: ImageRef
        """
        list.__init__(self, plate_trace)

        self._direction = image_dir
        self.plate = plate
        self.best_image = image_ref

    @property
    def direction(self):
        """
        :rtype: int
        """
        return self._direction

    @property
    def trace(self):
        """
        :rtype: list[int]
        """

        tr = []
        map(lambda item: tr.extend(list(item.plate_quad.center)), self)
        return tr

    @property
    def min_ts(self):
        return self[0].ts

    @property
    def max_ts(self):
        return self[-1].ts

    @staticmethod
    def deserialize(jnode):
        """
        :param jnode: TrTrack#/Lr
        :type jnode: dict
        :rtype: LrTrack
        """
        plate = Plate.deserialize_lr(jnode['Plate'])
        plate_dir = jnode['Dir']
        plate_trace = map(lambda item_jnode: LrTrackItem.deserialize(item_jnode),
                          jnode['PlateTrace'])

        # todo: find error in tracker/lpr and delete this workaround
        plate_trace.sort(key=lambda item: item.ts, reverse=False)

        image_jnode = jnode.get('Image')
        image_ref = JsonImRef.deserialize(image_jnode) if image_jnode else None

        return LrTrack(plate, plate_trace, plate_dir, image_ref)


class TrTrackItem:
    def __init__(self, ts, lr_item=None, ra_item=None):
        """
        :type ts: long
        :type lr_item: LrTrackItem | None
        :type ra_item: RaTrackItem | None
        """
        self.ts = ts
        self.lr_item = lr_item
        self.ra_item = ra_item

    def __lt__(self, other):
        """
        :type other: TrTrackItem
        """
        return self.ts < other.ts

    def get_plate_rect(self, lr_scene_w=None, lr_scene_h=None):
        """
        :type lr_scene_w: int | None
        :type lr_scene_h: int | None
        :rtype: RectF
        """
        # image_ts, fake_radar_plate_rect = best_ra_ts_rect
        if self.lr_item and lr_scene_w and lr_scene_h:
            return RectF(*self.lr_item.plate_quad.get_rel_bbox(lr_scene_w, lr_scene_h))

        elif self.ra_item:
            x = self.ra_item.x
            y = self.ra_item.y
            size = 0.05
            return RectF(x - size, y - size, x + size, y + size)

        else:
            raise NotImplementedError

    def get_plate_quad(self, lr_scene_w=None, lr_scene_h=None):
        """
        :type lr_scene_w: int | None
        :type lr_scene_h: int | None
        :rtype: QuadI
        """
        # image_ts, fake_radar_plate_rect = best_ra_ts_rect
        if self.lr_item:
            return self.lr_item.plate_quad

        elif self.ra_item and lr_scene_w and lr_scene_h:
            fx = self.ra_item.x * lr_scene_w
            fy = self.ra_item.y * lr_scene_h
            size = 0.05 * lr_scene_w

            fl = int(fx - size)
            ft = int(fy - size)
            fr = int(fx + size)
            fb = int(fy + size)
            return QuadI(fl, ft, fr, ft, fr, fb, fl, fb)

        else:
            raise NotImplementedError


class TrTrack(list):
    def __init__(self, trace):
        """
        :type trace: list[TrTrackItem]
        """
        list.__init__(self, trace)

    @property
    def min_ts(self):
        return self[0].ts

    @property
    def max_ts(self):
        return self[-1].ts

    @staticmethod
    def create(lr_track, ra_track):
        """
        :type lr_track: LrTrack | None
        :type ra_track: RaTrack | None
        :rtype: TrTrack
        """
        if lr_track and ra_track:
            lr_index = 0
            ra_index = 0
            max_lr_index = len(lr_track)
            max_ra_index = len(ra_track)

            trace = []
            while lr_index != max_lr_index and ra_index != max_ra_index:
                if lr_track[lr_index].ts == ra_track[ra_index].ts:
                    trace.append(TrTrackItem(lr_track[lr_index].ts, lr_track[lr_index], ra_track[ra_index]))
                    lr_index += 1
                    ra_index += 1

                elif lr_track[lr_index].ts < ra_track[ra_index].ts:
                    trace.append(TrTrackItem(lr_track[lr_index].ts, lr_item=lr_track[lr_index]))
                    lr_index += 1

                elif lr_track[lr_index].ts > ra_track[ra_index].ts:
                    trace.append(TrTrackItem(ra_track[ra_index].ts, ra_item=ra_track[ra_index]))
                    ra_index += 1

            if lr_index == max_lr_index:
                while ra_index < max_ra_index:
                    trace.append(TrTrackItem(ra_track[ra_index].ts, ra_item=ra_track[ra_index]))
                    ra_index += 1

            elif ra_index == max_ra_index:
                while lr_index < max_lr_index:
                    trace.append(TrTrackItem(lr_track[lr_index].ts, lr_item=lr_track[lr_index]))
                    lr_index += 1

        elif lr_track:
            trace = map(lambda item: TrTrackItem(item.ts, lr_item=item), lr_track)

        elif ra_track:
            trace = map(lambda item: TrTrackItem(item.ts, ra_item=item), ra_track)

        else:
            raise NotImplementedError

        return TrTrack(trace)

    def get_item_direction(self, item, max_lookup_depth=2):
        """
        :type item: TrTrackItem
        :type max_lookup_depth: float
        :param max_lookup_depth: in seconds
        :rtype: Dir
        """
        # find item index in self
        item_index = bisect.bisect_left(self, item)

        def _lr_direction():
            """
            :rtype: Dir
            """
            if item.lr_item is None:
                return Dir(Dir.UNKNOWN)

            prev_center = None

            # find first quad with center not in plate_quad but not far then max_lookup_depth seconds
            plate_quad = item.get_plate_quad()
            time_delta = long(max_lookup_depth * 1e6)
            for index in range(item_index, 0, -1):
                prev_item = self[index]  # type: TrTrackItem

                if item.ts - prev_item.ts > time_delta:
                    break

                if prev_item.lr_item is None:
                    continue

                prev_center_ = prev_item.get_plate_quad().center
                if prev_center_ in plate_quad:
                    continue

                prev_center = prev_center_
                break

            # define direction
            direction = Dir(Dir.UNKNOWN)
            if prev_center:
                plate_x, plate_y = plate_quad.center
                candidate_x, candidate_y = prev_center

                if candidate_y < plate_y:
                    direction = Dir(Dir.APPROACHING)

                else:
                    direction = Dir(Dir.RECEDING)

            return direction

        def _ra_direction():
            """
            :rtype: Dir
            """
            if item.ra_item is None:
                return Dir(Dir.UNKNOWN)

            # find first quad with center not in plate_quad but not far then max_lookup_depth seconds
            item_f = item.ra_item.raw_f
            prev_f = None
            time_delta = long(max_lookup_depth * 1e6)
            for index in range(item_index, 0, -1):
                prev_item = self[index]  # type: TrTrackItem
                if item.ts - prev_item.ts > time_delta:
                    break

                if prev_item.ra_item is None:
                    continue

                prev_f = prev_item.ra_item.raw_f
                break

            direction = Dir(Dir.UNKNOWN)
            if prev_f:
                if item_f > prev_f:
                    Dir(Dir.APPROACHING)
                else:
                    Dir(Dir.RECEDING)

            return direction

        ra_dir = _ra_direction()
        if ra_dir.dir != Dir.UNKNOWN:
            return ra_dir

        return _lr_direction()


class Track:
    def __init__(self, lr_track, ra_track, tr_track):
        """
        :type lr_track: LrTrack | None
        :type ra_track: RaTrack | None
        :type tr_track: TrTrack | None
        """
        self.uuid = str(uuid4())
        self.lr_track = lr_track
        self.ra_track = ra_track
        self.tr_track = tr_track

    @staticmethod
    def deserialize(tr_track_jnode):
        """
        :type tr_track_jnode: dict
        :rtype: Track
        """
        lr_track_jnode = tr_track_jnode.get('Lr')
        lr_track = LrTrack.deserialize(lr_track_jnode) if lr_track_jnode else None

        ra_track_jnode = tr_track_jnode.get('Ra')
        ra_track = RaTrack.deserialize(ra_track_jnode) if ra_track_jnode else None

        tr_track = TrTrack.create(lr_track, ra_track)

        track = Track(lr_track, ra_track, tr_track)
        return track
