import os
import time
import datetime

from edge.com.db import DbArgs
from edge import eventlog
from edge.types import mk_error_text, JnodeError

from ..types import OriginServiceArgs, BsServiceArgs
from ..types import RaasServiceArgs, ReactServiceArgs, EventServiceArgs, EventQueueArgs, TrainerServiceArgs
from ..types import StatServiceArgs, LaneDetectorServiceArgs, TlServiceArgs, DbServiceArgs, SystestServiceArgs
from ...detectors.parking.types import DetectorArgs as ParkingArgs
from ...detectors.red_light.types import DetectorArgs as RedLightArgs
from ...detectors.spot_speed.types import DetectorArgs as SpotSpeedArgs
from ...detectors.transit.types import DetectorArgs as TransitArgs
from ...detectors.wrong_way.types import DetectorArgs as WrongWayArgs
from ...model.types import Identity
from ...model.location import Location
from ...model.jnode.types import JsonMask


class JsonOriginServiceArgs(OriginServiceArgs):
    def __init__(self, *args, **kwargs):
        OriginServiceArgs.__init__(self, *args, **kwargs)

    @staticmethod
    def deserialize(origin_jnode):
        """
        :type origin_jnode: dict
        :rtype: JsonOriginServiceArgs
        :raise JnodeError:
        """
        # deserialize #/uuid
        uuid = origin_jnode.get('uuid')
        if not (type(uuid) in [unicode, str] and len(uuid) == 36):
            eventlog.log_d(mk_error_text('#/uuid'))
            raise JnodeError

        # deserialize #/time
        mk_time = origin_jnode.get('time')
        if not type(uuid) in [unicode, str]:
            eventlog.log_d(mk_error_text('#/time'))
            raise JnodeError

        location_jnode = origin_jnode.get('location', {})
        location = Location.deserialize(location_jnode)

        return JsonOriginServiceArgs(uuid, mk_time, location)


class JsonRaasServiceArgs(RaasServiceArgs):
    def __init__(self, *args, **kwargs):
        RaasServiceArgs.__init__(self, *args, **kwargs)

    @staticmethod
    def deserialize(raas_jnode):
        """
        :type raas_jnode: dict | None
        :rtype: JsonRaasServiceArgs
        :raise JnodeError:
        """
        if raas_jnode is None:
            return None

        # deserialize #/host
        raas_host = raas_jnode.get('host')
        if not (type(raas_host) in [str, unicode] and 1 <= len(raas_host) <= 16):
            eventlog.log_d(mk_error_text('#/host'))
            raise JnodeError

        # deserialize #/port
        raas_port = raas_jnode.get('port')
        if not (type(raas_port) in [str, unicode] and 1 <= len(raas_port) <= 16):
            eventlog.log_d(mk_error_text('#/port'))
            raise JnodeError

        # deserialize #/user
        raas_user = raas_jnode.get('user')
        if not (type(raas_user) in [str, unicode] and 1 <= len(raas_user) <= 16):
            eventlog.log_d(mk_error_text('#/user'))
            raise JnodeError

        # deserialize #/pass
        raas_pass = raas_jnode.get('pass')
        if not (type(raas_pass) in [str, unicode] and 1 <= len(raas_pass) <= 16):
            eventlog.log_d(mk_error_text('#/pass'))
            raise JnodeError

        return JsonRaasServiceArgs(raas_host, raas_port, raas_user, raas_pass)


class JsonEventServiceArgs(EventServiceArgs):
    def __init__(self, *args, **kwargs):
        EventServiceArgs.__init__(self, *args, **kwargs)

    @staticmethod
    def deserialize(fact_event_jnode):
        """
        :type fact_event_jnode: dict
        :rtype: JsonEventServiceArgs
        """
        max_file_count = fact_event_jnode.get('max-file-count', 15)
        if not (int == type(max_file_count) and 15 <= max_file_count <= 50):
            raise ValueError('main.conf#/event-store/max-file-count')

        json_queue_args_list = map(
            lambda endpoint: EventQueueArgs(endpoint, max_file_count - 5, max_file_count),
            fact_event_jnode['endpoints'])

        require_plate = fact_event_jnode.get('require-plate', True)
        return JsonEventServiceArgs(require_plate, json_queue_args_list)


class JsonReactServiceArgs(ReactServiceArgs):
    def __init__(self, *args, **kwargs):
        ReactServiceArgs.__init__(self, *args, **kwargs)

    @staticmethod
    def deserialize(main_conf_jnode):
        """
        :type main_conf_jnode: dict
        :rtype: JsonReactServiceArgs
        :raise JnodeError:
        """

        track_react_endpoint = main_conf_jnode['services']['react']

        return JsonReactServiceArgs(track_react_endpoint)


class JsonBsServiceArgs(BsServiceArgs):
    def __init__(self, *args, **kwargs):
        BsServiceArgs.__init__(self, *args, **kwargs)

    @staticmethod
    def deserialize(log_level, transit_args, wrong_way_args, spot_speed_args, parking_args, red_light_args):
        """
        :type log_level: str
        :type transit_args: TransitArgs
        :type wrong_way_args: list[WrongWayArgs]
        :type spot_speed_args: list[SpotSpeedArgs]
        :type parking_args: list[ParkingArgs]
        :type red_light_args: list[RedLightArgs]
        :rtype: JsonBsServiceArgs
        """
        from ...detectors.types import IDetector

        sender_to_chan_map = {
            Identity(IDetector.TYPE_TRANSIT, 0): {
                'args': transit_args.blob_store_args.vehicle_store,
                'plate': transit_args.blob_store_args.vehicle_store,
                'vehicle': transit_args.blob_store_args.vehicle_store,
                'thumb': transit_args.blob_store_args.vehicle_store,
                'scene': transit_args.blob_store_args.scene_store
            }
        }

        for args in wrong_way_args:
            sender_to_chan_map.update({args.identity: args.store})

        for args in spot_speed_args:
            sender_to_chan_map.update({args.identity: args.store})

        for args in parking_args:
            sender_to_chan_map.update({args.identity: args.store})

        for args in red_light_args:
            sender_to_chan_map.update({args.identity: args.store})

        return JsonBsServiceArgs(sender_to_chan_map, log_level)


class JsonTrainerServiceArgs(TrainerServiceArgs):
    def __init__(self, *args, **kwargs):
        TrainerServiceArgs.__init__(self, *args, **kwargs)

    @staticmethod
    def deserialize(trainer):
        """
        :type trainer: dict
        :rtype: JsonTrainerServiceArgs
        :raise JnodeError:
        """
        try:
            db_directory = trainer['directory']
            max_file_count = trainer.get('max-file-count', 1000)
            min_file_count = max_file_count - 10
            time_min_max = trainer['limits'].get('time-min-max', [])
            lanes = trainer['limits'].get('lanes', [])

            min_time = datetime.time(hour=0, minute=0, second=0)
            max_time = datetime.time(hour=23, minute=59, second=59)
            if len(time_min_max) == 2:
                raw_min_time, raw_max_time = time_min_max
                min_struct_time = time.strptime(raw_min_time, "%H:%M")
                max_struct_time = time.strptime(raw_max_time, "%H:%M")

                min_time = datetime.time(min_struct_time.tm_hour, min_struct_time.tm_min)
                max_time = datetime.time(max_struct_time.tm_hour, max_struct_time.tm_min)

            return JsonTrainerServiceArgs(db_directory, lanes, min_time, max_time, min_file_count, max_file_count)
        except KeyError:
            eventlog.log_d('fail: plate_absent_db jnode have invalid keys')

        raise JnodeError


class JsonStatServiceArgs(StatServiceArgs):
    def __init__(self, *args, **kwargs):
        StatServiceArgs.__init__(self, *args, **kwargs)

    @staticmethod
    def deserialize(monitoring_jnode):
        """
        :type monitoring_jnode: dict
        :rtype: StatServiceArgs
        :raise JnodeError:
        """
        stat_filepath = monitoring_jnode.get('path')
        if not (type(stat_filepath) in [str, unicode] and len(stat_filepath) < 64 and os.path.isabs(stat_filepath)):
            eventlog.log_d(mk_error_text('#/path'))
            raise JnodeError

        stat_tout_sec = monitoring_jnode.get('tout')
        if not (int == type(stat_tout_sec) and 0 <= stat_tout_sec <= 255):
            eventlog.log_d(mk_error_text('#/tout'))
            raise JnodeError

        return JsonStatServiceArgs(stat_filepath, stat_tout_sec)


class JsonLaneDetectorServiceArgs(LaneDetectorServiceArgs):
    def __init__(self, *args, **kwargs):
        LaneDetectorServiceArgs.__init__(self, *args, **kwargs)

    @staticmethod
    def deserialize(raw_road_jnode):
        """
        :type raw_road_jnode: dict
        :rtype: LaneDetectorArgs
        :raise JnodeError:
        """
        if not len(raw_road_jnode):
            eventlog.log_d('warning: no road.conf, making fake LaneDetector args')
            return JsonLaneDetectorServiceArgs([], (14, 14))

        road_jnode = raw_road_jnode.get('road')
        if not type(road_jnode) in [dict]:
            eventlog.log_d(mk_error_text('#/road'))
            raise JnodeError

        # deserialize #/mask/resolution
        resolution_jnode = raw_road_jnode.get('scene-wh')
        if not resolution_jnode or len(resolution_jnode) != 2:
            eventlog.log_d(mk_error_text('#/scene-wh'))
            raise JnodeError

        im_width, im_height = resolution_jnode
        if not (type(im_width) in [int] and 0 <= im_width <= 65535):
            eventlog.log_d(mk_error_text('#/mask/scene-wh/0'))
            raise JnodeError

        if not (type(im_height) in [int] and 0 <= im_height <= 65535):
            eventlog.log_d(mk_error_text('#/mask/scene-wh/1'))
            raise JnodeError

        # deserialize #/mask/lanes
        lanes_jnode = road_jnode.get('lanes', [])
        if list != type(lanes_jnode):
            eventlog.log_d(mk_error_text('#/mask/lanes'))
            raise JnodeError

        id_to_lane_list = map(
            lambda lane_jnode: (unicode(lane_jnode['name']), JsonMask.deserialize(lane_jnode['mask'])),
            lanes_jnode)

        return JsonLaneDetectorServiceArgs(id_to_lane_list, (im_width, im_height))


class JsonTlServiceArgs(TlServiceArgs):
    def __init__(self, *args, **kwargs):
        TlServiceArgs.__init__(self, *args, **kwargs)

    @staticmethod
    def deserialize(jnode):
        """
        :type jnode: dict | None
        :rtype: JsonTlServiceArgs | None
        :raise JnodeError:
        """
        if jnode is None:
            return None

        if len(jnode) == 2:
            endpoint, tl_ident = jnode
            return JsonTlServiceArgs(endpoint, tl_ident)


class JsonDbServiceArgs(DbServiceArgs):
    def __init__(self, *args, **kwargs):
        DbServiceArgs.__init__(self, *args, **kwargs)

    @staticmethod
    def deserialize(origin_id, jnode):
        """
        :type origin_id: int
        :type jnode: dict
        :rtype: JsonDbServiceArgs
        :raise JnodeError:
        """
        db_jnode = jnode.get('db')
        if not db_jnode:
            eventlog.log_e('fail: param #/db is absent')
            raise JnodeError

        db_args = DbArgs.deserialize(db_jnode)
        return JsonDbServiceArgs(origin_id, db_args)


class JsonSystestServiceArgs(SystestServiceArgs):
    def __init__(self, *args, **kwargs):
        SystestServiceArgs.__init__(self, *args, **kwargs)

    @staticmethod
    def deserialize(main_conf_jnode):
        """
        :type main_conf_jnode: dict
        :rtype: JsonSystestServiceArgs
        :raise JnodeError:
        """

        systest_conf_filepath = main_conf_jnode['services'].get('systest')
        return JsonSystestServiceArgs(systest_conf_filepath)
