import time
import datetime
import ujson
import struct

from edge.com.fs import silent_mkdir_r
from edge.queue import ImageFileQueue
from edge_bbox.types import RectF
from edge.image.types import MemBlob
from edge.service.tz import ITzService

from .types import TrainerServiceArgs
from ..model.types import BaseEvent, BestWaypoint


def init_service(twisted_reactor, args, tz_service):
    """
    :type twisted_reactor: twisted.internet.interfaces.IReactorTCP
    :type args: TrainerServiceArgs | None
    :type tz_service: ITzService
    :rtype: IReactService
    """
    return _TrainerService(twisted_reactor, args, tz_service) if args else _StubTrainerService()


# noinspection PyUnusedLocal
def fini_service(service):
    """
    :type service: ITrainerService | _TrainerService
    """
    pass


class ITrainerService:
    def __init__(self):
        pass

    def collect(self, event_args, blob, ra_rect):
        """
        :type event_args: BaseEvent,
        :type ra_rect: RectF
        :type blob: MemBlob
        """
        raise NotImplementedError


class _StubTrainerService(ITrainerService):
    def __init__(self):
        ITrainerService.__init__(self)

    def collect(self, *args, **kwargs):
        pass


class _TrainerService(ITrainerService):
    def __init__(self, twisted_reactor,  args, tz_service):
        """
        :type twisted_reactor: twisted.internet.interfaces.IReactorTCP
        :type args: TrainerServiceArgs
        :type tz_service: ITzService
        """
        ITrainerService.__init__(self)
        self._root_dirpath = args.root_dir
        silent_mkdir_r(self._root_dirpath)

        self._tz_service = tz_service
        self._lanes = args.lanes
        self._min_time = args.min_time
        self._max_time = args.max_time
        self._jpeg_queue = ImageFileQueue(
            twisted_reactor, 'JPEG', args.root_dir,
            args.min_file_count, args.max_file_count)

    def collect(self, event_args, blob, ra_rect):
        """
        :type event_args: BaseEvent
        :type blob: MemBlob
        :type ra_rect: RectF
        """
        # define best-waypoint
        best_waypoint = filter(
            lambda point: point.name == point.NAME_BEST,
            event_args.measurements[0].waypoints)[0]  # type: BestWaypoint

        #
        if not self._lanes or best_waypoint.location.lane in self._lanes:
            image_struct_time = time.gmtime(blob.min_ts / 1000000)
            image_time = datetime.time(image_struct_time.tm_hour, image_struct_time.tm_min)

            if self._min_time < image_time < self._max_time:
                # add event jnode to the end of jpeg
                event_jnode = event_args.serialize(self._tz_service)
                event_str = ujson.dumps(event_jnode)
                # add event jnode size after jnode
                event_size = struct.pack('I', len(event_str))

                self._jpeg_queue.add_blob_image(blob.file_data + event_str + event_size, blob.min_ts)
