from edge import eventlog
from edge.types import ImageRef

from .services.react import SceneReactProducer, TrackReactProducer
from .services.event import SceneEventConsumer, TrackEventConsumer
from .services.stats import StatCounters

from _args import KernelArgs, LrUpscaleArgs
from _match_store import LrRaMatchStore
from _models import LrScene, LrTrack, RaScene, RaTrack, TrScene, TrTrack
from _ra_track_store import RaTrackStore


class BaseKernel:
    def __init__(self,
                 scene_react_producer,
                 track_react_producer,
                 scene_event_consumers,
                 track_event_consumer,
                 lr_ra_match_store,
                 ra_track_store,
                 min_ra_track_len,
                 lr_upscale):
        """
        :type scene_react_producer: SceneReactProducer
        :type track_react_producer: TrackReactProducer
        :type scene_event_consumers: list[SceneEventConsumer]
        :type track_event_consumer: TrackEventConsumer
        :type lr_ra_match_store: LrRaMatchStore
        :type ra_track_store: RaTrackStore
        :type min_ra_track_len: int
        :type lr_upscale: LrUpscaleArgs
        """
        self._min_ra_track_len = min_ra_track_len

        self._scene_react_producer = scene_react_producer
        self._scene_react_producer.set_handler(self._handle_scene_react)

        self._track_react_producer = track_react_producer
        self._track_react_producer.set_handler(self._handle_track_react)

        self._lr_ra_match_store = lr_ra_match_store
        self._ra_track_store = ra_track_store
        self._ra_track_store.set_tout_handler(self._handle_ra_track_tout)

        self._scene_event_consumers = scene_event_consumers
        self._track_event_consumer = track_event_consumer

        self._lr_upscale = lr_upscale

    def _handle_scene_react(self, im_ref, lr_scene, ra_scene):
        """
        :type im_ref: ImageRef | None
        :type lr_scene: LrScene | None
        :type ra_scene: RaScene | None
        """
        # Update _ra_track_store
        if ra_scene:
            self._ra_track_store.add_ra_scene(ra_scene)

        # Create the tr_scene from an avaliable lr_scene and ra_scene.
        if lr_scene and ra_scene:
            tr_scene = TrScene.create_lr_ra(
                self._lr_upscale, lr_scene.scene, ra_scene.scene,
                lr_scene.im_ref.width, lr_scene.im_ref.height)
            self._lr_ra_match_store.add(lr_scene.source_id, tr_scene.tr_items)

        elif lr_scene:
            tr_scene = TrScene.create_lr(self._lr_upscale, lr_scene.scene)

        elif ra_scene:
            tr_scene = TrScene.create_ra(self._lr_upscale, ra_scene.scene)

        else:
            tr_scene = TrScene.create_empty(self._lr_upscale)

        if im_ref:
            for consumer in self._scene_event_consumers:
                consumer.consume_scene(im_ref, tr_scene)

    def _handle_track_react(self, lr_track):
        """
        :type lr_track: LrTrack
        """
        ra_track = None
        ra_oids_desc = self._lr_ra_match_store.get_ra_oids(lr_track.source_id, lr_track.oids)
        for ra_oid in ra_oids_desc:
            ra_track = self._ra_track_store.try_get_ra_track(ra_oid)
            if ra_track:
                break

        if ra_track is None:
            tr_track = TrTrack(lr_track.src_jnode, None)

        elif len(ra_track) >= self._min_ra_track_len:
            self._ra_track_store.rm_ra_track(ra_track.oid)
            tr_track = TrTrack(lr_track.src_jnode, ra_track)

        else:
            tr_track = TrTrack(lr_track.src_jnode, None)

        self._track_event_consumer.consume_track(tr_track, plate_text=lr_track.plate_text)

    def _handle_ra_track_tout(self, ra_track):
        """
        :type ra_track: RaTrack
        """
        if len(ra_track) >= self._min_ra_track_len:
            tr_track = TrTrack(None, ra_track)

            # noinspection PyBroadException
            try:
                self._track_event_consumer.consume_track(tr_track)
            except Exception:
                eventlog.log_x()


class Kernel(BaseKernel):
    def __init__(self, twisted_reactor, args):
        """
        :type args: KernelArgs
        """
        stats = StatCounters(
            twisted_reactor,
            scene_consumer_count=len(args.scene_event_args_list),
            stat_filepath=args.stat_counters_args.stat_filepath,
            stat_tout_sec=args.stat_counters_args.stat_tout_sec)

        track_react_producer = TrackReactProducer(
            twisted_reactor, args.track_react_args, stats.track_react)

        scene_react_producer = SceneReactProducer(
            twisted_reactor, args.scene_react_args, stats.scene_react)

        ident = 0
        scene_event_consumers = list()
        for scene_event_args in args.scene_event_args_list:
            consumer = SceneEventConsumer(ident, scene_event_args, stats.scene_event)
            scene_event_consumers.append(consumer)
            ident += 1

        track_event_consumer = TrackEventConsumer(
            twisted_reactor,
            args.track_event_args,
            stats.track_event,
            args.image_store_args)

        ra_track_store = RaTrackStore(twisted_reactor, args.ra_track_store_tout)
        lr_ra_match_store = LrRaMatchStore(twisted_reactor, args.lr_ra_oid_match_store_tout)

        BaseKernel.__init__(
            self,
            scene_react_producer,
            track_react_producer,
            scene_event_consumers,
            track_event_consumer,
            lr_ra_match_store,
            ra_track_store,
            args.min_ra_track_len,
            args.lr_upscale_args)

        self._twisted_reactor = twisted_reactor
        self._stat_counters = stats

    def run(self):
        self._twisted_reactor.run()
